(*
Copyright 2012 René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Bounded_Increase_Impl
imports 
  Bounded_Increase 
  RTrancl 
  QDP_Framework_Impl
  Q_Restricted_Rewriting_Impl
  Non_Inf_Order_Impl
  Dependency_Graph_Impl
  Linear_Orders
  Orthogonality_Impl
  Innermost_Usable_Rules_Impl
begin

fun rel_dep_prod :: "bool \<Rightarrow> dependance \<Rightarrow> bool list" (infix "***" 52) where
  "_ *** Ignore = []"
| "b *** Increase = [b]"
| "b *** Decrease = [\<not> b]"
| "_ *** Wild = [True,False]"

fun usable_rules_gen1 :: "'f dep \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)term \<times> bool \<Rightarrow> (('f,'v)term \<times> bool)list" where 
  [code]: "usable_rules_gen1 \<pi> R (Var _,_) = []"
| "usable_rules_gen1 \<pi> R (Fun f ts, b) =  [ (r,b) . (l,r) <- R, compat_root l (Fun f ts)] @
     [(ts ! i, d). i <- [0 ..< length ts], d <- b *** \<pi> (f,length ts) i]"

lemma [code]: "usable_rules_gen1 \<pi> R (Fun f ts, b) =  (let n = length ts in [ (r,b) . (l,r) <- R, compat_root l (Fun f ts)] @
     [(ts ! i, d). i <- [0 ..< n], d <- b *** \<pi> (f,n) i])" by (simp add: Let_def)

fun usable_rules_gen2 :: "('f,'v)rules \<Rightarrow> ('f,'v)term \<times> bool \<Rightarrow> ('f,'v)rules" where
  "usable_rules_gen2 R (t, b) = (let RR = [(l,r) . (l,r) <- R, compat_root l t] in (if b then RR else map (\<lambda> (l,r). (r,l)) RR))"

definition usable_rules_pre_gen :: "'f dep \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules" where
  "usable_rules_pre_gen \<pi> R P \<equiv> remdups (concat (map (usable_rules_gen2 R) 
    (mk_rtrancl_list (op =) (usable_rules_gen1 \<pi> R) (map (\<lambda> (s,t). (t,True)) P))))"

lemma (in fixed_trs_dep) usable_rules_pre_gen: assumes RR: "set RR = R"
  shows "set (usable_rules_pre_gen \<pi> RR P) = gen_usable_rules_pairs (set P)"
proof -
  let ?U1 = "usable_rules_gen1 \<pi> RR"
  let ?U2 = "usable_rules_gen2 RR"
  {
    fix t b
    have "{ba. ((t, b), ba) \<in> {(a, b). b \<in> set (?U1 a)}^*} \<subseteq> {(s,d) | s u d. u \<in> set (t # map snd RR) \<and> u \<unrhd> s}" (is "?A \<subseteq> ?B")
    proof
      fix s c
      assume "(s,c) \<in> ?A"
      hence "((t, b), s, c) \<in> {(a, b). b \<in> set (?U1 a)}^*" by simp      
      thus "(s,c) \<in> ?B"
      proof (induct)
        case base
        show ?case by auto
      next
        case (step sc ud)
        obtain s c where sc: "sc = (s,c)" by force
        obtain u d where ud: "ud = (u,d)" by force
        from step(2)[unfolded sc ud]
        have mem: "(u, d) \<in> set (?U1 (s, c))" by auto
        from mem obtain f ss where s: "s = Fun f ss" by (cases s, auto)
        from mem[unfolded s]
        have  "(\<exists>l. (l, u) \<in> set RR \<and> compat_root l (Fun f ss) \<and> d = c) \<or>
          (\<exists> i. i <length ss \<and> (u, d) \<in> Pair (ss ! i) ` set (c *** \<pi> (f, length ss) i))" (is "?l \<or> ?i") by force
        thus ?case
        proof
          assume ?l
          then obtain l where "(l, u) \<in> set RR" by auto
          hence "u \<in> set (t # map snd RR) \<and> u \<unrhd> u" by force
          thus ?thesis unfolding ud by blast
        next
          assume ?i
          then obtain i where "i < length ss" and u: "u = ss ! i" by auto
          hence su: "s \<unrhd> u" unfolding s by auto
          from step(3)[unfolded sc] obtain v where v: "v \<in> set (t # map snd RR)" and vs: "v \<unrhd> s" by auto
          from supteq_trans[OF vs su] v show ?thesis unfolding ud by auto
        qed
      qed
    qed
    also have "... = set [(s,d) . u <- t # map snd RR, s <- supteq_list u, d <- [True,False]]" (is "_ = ?B")
      by auto
    finally have "?A \<subseteq> ?B" .
    from finite_subset[OF this finite_set] have "finite ?A" .
  } note finite = this          
  interpret relation_subsumption_list ?U1 "op ="
    by (unfold_locales, insert finite, auto)
  let ?UU1 = "{(a, b). b \<in> set (?U1 a)}"
  have id: "set (usable_rules_pre_gen \<pi> RR P) = 
    \<Union> ((\<lambda> tc. set (?U2 tc)) ` ?UU1^* `` {(t,True) | t. t \<in> rhss (set P)})" (is "?A1 = ?A2")
    unfolding usable_rules_pre_gen_def by (force simp: mk_rtrancl_list[unfolded mk_rtrancl_no_subsumption[OF refl]])
  have "gen_usable_rules_pairs (set P) = \<Union>gen_usable_rules ` rhss (set P)" (is "?B1 = ?B2")
    unfolding gen_usable_rules_pairs_def ..
  also have "?B2 = ?A2" 
  proof 
    show "?A2 \<subseteq> ?B2"
    proof (rule subsetI)    
      def dc \<equiv> "\<lambda> c. if c then Increase else Decrease"
      {
        fix t s c
        assume "((t,True),(s,c)) \<in> ?UU1^*"
        from rtrancl_imp_UN_relpow[OF this] obtain n where 
          "((t,True),(s,c)) \<in> ?UU1 ^^ n" by auto
        hence "gen_usable_rules s ^^^ dc c \<subseteq> gen_usable_rules t"
        proof (induction n arbitrary: s c)
          case 0
          thus ?case unfolding dc_def by auto
        next
          case (Suc n s c)
          from Suc.prems[unfolded relpow.simps]
          obtain u d where steps: "((t,True),(u,d)) \<in> ?UU1 ^^ n" and step: "(s,c) \<in> set (?U1 (u,d))" by force
          from step obtain f us where u: "u = Fun f us" by (cases u, auto)
          from step[unfolded u]
          have "(\<exists>l. (l,s) \<in> R \<and> compat_root l (Fun f us) \<and> c = d) \<or>
                (\<exists>i < length us. s = us ! i \<and> c \<in> set (d *** \<pi> (f, length us) i))" by (force simp: RR)
          then obtain l i where "(l,s) \<in> R \<and> compat_root l (Fun f us) \<and> c = d \<or> 
            i < length us \<and> s = us ! i \<and> c \<in> set (d *** \<pi> (f, length us) i)" (is "?l \<or> ?i") by blast
          hence "gen_usable_rules s ^^^ dc c \<subseteq> gen_usable_rules u ^^^ dc d" (is "?left \<subseteq> ?right")
          proof
            assume l: ?l
            then obtain ls where "l = Fun f ls" and "length ls = length us" and "(Fun f ls,s) \<in> R" and "c = d" by (cases l, auto)
            hence "gen_usable_rules s \<subseteq> \<Union> { gen_usable_rules r | ls r. (Fun f ls,r) \<in> R \<and> length ls = length us}" by auto
            also have "... \<subseteq> gen_usable_rules u" unfolding u gen_usable_rules_Fun by blast
            finally show ?thesis unfolding `c = d`
              by (rule rel_dep_mono)
          next
            assume i: ?i
            {
              fix s :: "('f,'v)trs" and b1 b2 d
              assume "b1 \<in> set (b2 *** d)"
              hence "s ^^^ dc b1 \<subseteq> (s ^^^ d) ^^^ dc b2" unfolding dc_def
                by (cases b1, cases b2, cases d, auto, cases d, auto, cases d, auto, cases d, auto)
            } note dep_prod = this
            have "gen_usable_rules s ^^^ dc c \<subseteq>
              (gen_usable_rules s ^^^ \<pi> (f,length us) i) ^^^ dc d" 
              by (rule dep_prod, insert i, auto)
            also have "... \<subseteq> gen_usable_rules u ^^^ dc d"
              by (rule rel_dep_mono, unfold u gen_usable_rules_Fun, insert i, auto)
            finally show ?thesis .
          qed
          also have "... \<subseteq> gen_usable_rules t" by (rule Suc.IH[OF steps])
          finally show ?case .
        qed
      } note main = this
      fix lr
      assume "lr \<in> ?A2"      
      then obtain t s c where U2: "lr \<in> set (?U2 (s,c))" and U1: "((t,True),(s,c)) \<in> {(a, b). b \<in> set (?U1 a)}^*" 
        and P: "t \<in> rhss (set P)" by force
      obtain l r where lr: "lr = (l,r)" by force
      have "(l,r) \<in> gen_usable_rules s ^^^ dc c"
      proof (cases c)
        case True
        with U2 lr have "(l,r) \<in> set RR" and "compat_root l s" by auto
        from in_R[OF this[unfolded RR]] True show ?thesis unfolding gen_usable_rules_def dc_def by auto
      next
        case False
        with U2 lr have "(r,l) \<in> set RR" and "compat_root r s" by auto
        from in_R[OF this[unfolded RR]] False show ?thesis unfolding gen_usable_rules_def dc_def by auto
      qed
      also have "gen_usable_rules s ^^^ dc c \<subseteq> ?B2" using main[OF U1] P by auto
      finally
      show "lr \<in> ?B2" unfolding lr .
    qed
  next
    show "?B2 \<subseteq> ?A2"
    proof
      fix l r
      assume "(l,r) \<in> ?B2"
      then obtain t where gen: "gen_usable_rule t (l,r)" and t: "t \<in> rhss (set P)"
        using gen_usable_rules_def by blast
      from gen have "\<exists> s d. ((t,True),(s,d)) \<in> ?UU1^* \<and> (l,r) \<in> set (?U2 (s,d))"
      proof (induct)
        case (in_R l r t)
        show ?case
          by (rule exI[of _ t], rule exI[of _ True], insert in_R, auto simp: RR)
      next
        case (in_U l r t lr')
        from in_U(4) obtain s d where steps: "((r,True),(s,d)) \<in> ?UU1^*" and lr': "lr' \<in> set (?U2 (s,d))" by blast
        from in_U(1-2) have "((t,True),(r,True)) \<in> ?UU1" by (cases t, auto simp: RR)
        from this steps have "((t,True),(s,d)) \<in> ?UU1^*" by (rule converse_rtrancl_into_rtrancl)
        with lr' show ?case by auto
      next
        case (in_arg i ts lr lr' f)
        from in_arg(3) obtain s d where steps: "((ts ! i,True),(s,d)) \<in> ?UU1^*" and mem: "lr \<in> set (?U2 (s,d))" by blast
        obtain l r where lr: "lr = (l,r)" by force
        from mem lr have imem: "(r,l) \<in> set (?U2 (s,\<not>d))" by (auto simp: Let_def)
        {
          fix t b s c
          assume "((t,b),(s,c)) \<in> ?UU1^*"
          from rtrancl_imp_UN_relpow[OF this] obtain n 
            where "((t,b),(s,c)) \<in> ?UU1^^n" by blast
          hence "((t,\<not> b),(s,\<not>c)) \<in> ?UU1^*"
          proof (induction n arbitrary: t b s c)
            case 0 thus ?case by auto
          next
            case (Suc n t b s c)
            from Suc.prems obtain u d 
              where steps: "((t,b),(u,d)) \<in> ?UU1^^n" and step: "(s,c) \<in> set (?U1 (u,d))" by force
            from Suc.IH[OF steps] have IH: "((t,\<not>b), (u,\<not>d)) \<in> ?UU1^*" .
            {              
              from step obtain f us where u: "u = Fun f us" by (cases u, auto)
              with step have "(\<exists>a\<in>set RR. \<exists>l. (l, s) = a \<and> compat_root l (Fun f us) \<and> c = d) \<or>
                (\<exists>i < length us. s = us ! i \<and> c \<in> set (d *** \<pi> (f, length us) i))" (is "?l \<or> ?i") by auto
              hence "(s,\<not>c) \<in> set (?U1 (u,\<not>d))"
              proof
                assume ?l thus ?thesis unfolding u by auto
              next
                assume ?i
                then obtain i where i: "i < length us" and s: "s = us ! i" and c: "c \<in> set (d *** \<pi> (f, length us) i)" by auto
                let ?pi = "\<pi> (f,length us) i"
                from c have "(\<not> c) \<in> set ((\<not> d) *** ?pi)" by (cases ?pi, auto)
                with i s show ?thesis unfolding u by auto
              qed
            }
            with IH have "((t,\<not>b),(s,\<not>c)) \<in> ?UU1 ^* O ?UU1" by auto
            thus ?case by regexp
          qed
        } note invert = this
        from in_arg(1) have i: "i < length ts" .
        from in_arg(4) have lr': "lr' \<in> {lr} ^^^ \<pi> (f, length ts) i" .
        from invert[OF steps] have "((ts ! i, False), s, \<not> d) \<in> ?UU1^*" by simp
        note isteps = converse_rtrancl_into_rtrancl[OF _ this]
        note steps = converse_rtrancl_into_rtrancl[OF _ steps]
        show ?case 
        proof (cases "\<pi> (f,length ts) i")
          case Ignore
          with lr' show ?thesis by auto
        next
          case Increase
          with lr' have lr': "lr' = lr" by auto
          from Increase i have "((Fun f ts, True), (ts ! i, True)) \<in> ?UU1" by force
          from steps[OF this] show ?thesis using mem lr' by auto
        next
          case Decrease
          with lr' lr have lr': "lr' = (r,l)" by auto
          from Decrease i have "((Fun f ts, True), (ts ! i, False)) \<in> ?UU1" by force
          from isteps[OF this] imem lr' show ?thesis by force
        next
          case Wild
          from lr' have "lr' = (l,r) \<or> lr' = (r,l)" unfolding lr Wild by auto
          thus ?thesis
          proof
            assume lr': "lr' = (l,r)"
            from Wild i have "((Fun f ts, True), (ts ! i, True)) \<in> ?UU1" by force
            from steps[OF this] show ?thesis using mem lr lr' by auto
          next
            assume lr': "lr' = (r,l)"
            from Wild i have "((Fun f ts, True), (ts ! i, False)) \<in> ?UU1" by force
            from isteps[OF this] imem lr' show ?thesis by force
          qed
        qed
      qed
      thus "(l,r) \<in> ?A2" using t by blast
    qed
  qed
  finally show ?thesis unfolding id by simp
qed

fun compat_root' where 
  "compat_root' _ None = False"
| "compat_root' (Var _) _ = False"
| "compat_root' fls gn = (root fls = gn)"

lemma compat_root'[simp]: "compat_root' l (root t) = compat_root l t"
  by (cases l, (cases t, auto)+)

fun usable_rules_gen2' :: "('f,'v)rules \<Rightarrow> ('f \<times> nat)option \<times> bool \<Rightarrow> ('f,'v)rules" where
  "usable_rules_gen2' R (fn, b) = (let RR = [(l,r) . (l,r) <- R, compat_root' l fn] in (if b then RR else map (\<lambda> (l,r). (r,l)) RR))"

definition usable_rules_gen :: "'f dep \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules" where
  "usable_rules_gen \<pi> R P \<equiv> concat (map (usable_rules_gen2' R) 
    (remdups (map (\<lambda> (t,b). (root t,b)) (mk_rtrancl_list (op =) (usable_rules_gen1 \<pi> R) (map (\<lambda> (s,t). (t,True)) P)))))"

lemma (in fixed_trs_dep) usable_rules_gen: assumes RR: "set RR = R"
  shows "set (usable_rules_gen \<pi> RR P) = gen_usable_rules_pairs (set P)"
proof -
  {
    fix t b
    have "set (usable_rules_gen2' RR (root t,b)) = set (usable_rules_gen2 RR (t,b))"
      by (auto simp: Let_def)
  }
  thus ?thesis
    unfolding usable_rules_pre_gen[OF RR, symmetric]
    unfolding usable_rules_gen_def usable_rules_pre_gen_def
    by force
qed

(* note, that usable_rules_pre_gen = usable_rules_gen does not hold, since the order may be different.
   hence, it is not possible to install a code-equation for usable_rules_pre_gen *)

fun deep_normalize_cc' :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint" where
  "deep_normalize_cc' (CC_impl [] c) = deep_normalize_cc' c"
| "deep_normalize_cc' (CC_impl cs c) = CC_impl (map deep_normalize_cc' cs) (deep_normalize_cc' c)"
| "deep_normalize_cc' (CC_cond s c) = (CC_cond s c)"
| "deep_normalize_cc' (CC_all s c) = (CC_all s (deep_normalize_cc' c))"
| "deep_normalize_cc' (CC_rewr s c) = (CC_rewr s c)"

definition deep_normalize_cc :: "('v list \<Rightarrow> 'v) \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint" where
  "deep_normalize_cc fresh c \<equiv> fresh.normalize_alpha fresh (deep_normalize_cc' c)"

lemma (in fixed_trs_dep_order_fresh) deep_normalize_cc: "\<sigma> \<Turnstile> deep_normalize_cc fresh c = \<sigma> \<Turnstile> c"
  unfolding deep_normalize_cc_def normalize_alpha
  by (induct c arbitrary: \<sigma> rule: deep_normalize_cc'.induct, auto)

function(sequential) check_subsumes' :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> bool" where
  "check_subsumes' (CC_impl cs c) (CC_impl ds d) = (check_subsumes' c d \<and> (\<forall> c \<in> set cs. \<exists> d \<in> set ds. check_subsumes' d c))"
| "check_subsumes' c (CC_impl ds d) = check_subsumes' c d"
| "check_subsumes' (CC_all x c) (CC_all y d) = (x = y \<and> check_subsumes' c d)"
| "check_subsumes' c d = (c = d)"
  by pat_completeness auto

termination
proof 
  show "wf (measure (\<lambda> (x,y). size x + size y))" by auto
next
  fix x y c d :: "('f,'v)cond_constraint" and ds cs :: "('f,'v)cond_constraint list"
  assume x: "x \<in> set ds" and y: "y \<in> set cs" 
  from size_simps(1)[OF x] size_simps(1)[OF y]
  show "((x, y), CC_impl cs c, CC_impl ds d) \<in> measure (\<lambda> (x,y). size x + size y)"
    by auto
qed auto

lemma (in fixed_trs_dep_order_fresh) check_subsumes': assumes "check_subsumes' c d"
  shows "c \<longrightarrow>cc d"
proof 
  fix \<sigma>
  assume "\<sigma> \<Turnstile> c"
  with assms show "\<sigma> \<Turnstile> d"
  proof (induct c d arbitrary: \<sigma> rule: check_subsumes'.induct)
    case (1 cs c ds d \<sigma>)
    show ?case unfolding cc_models.simps
    proof
      assume ds: "Ball (set ds) (op \<Turnstile> \<sigma>)"
      from 1(3) have subs: "check_subsumes' c d" by auto
      have "Ball (set cs) (op \<Turnstile> \<sigma>)"
      proof
        fix c
        assume c: "c \<in> set cs"
        with 1(3) obtain d where "d \<in> set ds" and "check_subsumes' d c" by auto
        with 1(2)[OF c this] ds show "\<sigma> \<Turnstile> c" by auto
      qed
      with 1(4) have "\<sigma> \<Turnstile> c" by simp
      from 1(1)[OF subs this] show "\<sigma> \<Turnstile> d" .
    qed
  qed auto
qed

definition check_subsumes :: "('v list \<Rightarrow> 'v) \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> bool" where
  "check_subsumes fresh c d \<equiv> let n = deep_normalize_cc fresh
    in check_subsumes' (n c) (n d)"

lemma (in fixed_trs_dep_order_fresh) check_subsumes: assumes "check_subsumes fresh c d"
  shows "c \<longrightarrow>cc d"
proof -
  from check_subsumes'[OF assms[unfolded check_subsumes_def Let_def]]
  have deep: "deep_normalize_cc fresh c \<longrightarrow>cc deep_normalize_cc fresh d" .
  show ?thesis
  proof
    fix \<sigma>
    assume model: "\<sigma> \<Turnstile> c"
    show "\<sigma> \<Turnstile> d" using cc_implies_substE[OF deep, unfolded deep_normalize_cc, OF model] .
  qed
qed  

definition initial_conditions_gen_impl where
  "initial_conditions_gen_impl p bef_len aft_len P st \<equiv>
  let pairs = \<lambda>n. generate_lists n P; all = [bef @ st # aft. bef <- pairs bef_len, aft <- pairs aft_len]
  in filter (\<lambda> bef_st_aft. \<forall>i<bef_len + aft_len. p (bef_st_aft ! i) (bef_st_aft ! Suc i)) all"

lemma initial_conditions_gen_impl[simp]: "set (initial_conditions_gen_impl p bef aft P st) = 
  initial_conditions_gen p bef aft (set P) st" 
  unfolding initial_conditions_gen_def initial_conditions_gen_impl_def Let_def by auto

abbreviation fresh_xx where "fresh_xx \<equiv> fresh_string ''xx''"

definition check_constraint_present :: "('dpp,'f :: {show,key},string)dpp_ops \<Rightarrow> 'dpp \<Rightarrow>
  'f \<Rightarrow> ('f,string)rules \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 
  (('f,string)cond_constraint \<times> ('f,string)rules) list \<Rightarrow>
  (condition_type \<Rightarrow> ('f,string)rule \<Rightarrow> shows check)" where
  "check_constraint_present I dpp constant p bef aft ccs \<equiv> let
    edg = is_iedg_edge_dpp I dpp;
    init_conds = initial_conditions_gen_impl (\<lambda> st uv. edg st (fst uv)) bef aft p
  in (\<lambda> ct st. check_allm (\<lambda> sts. check (\<exists> (c,uvs) \<in> set ccs. disjoint_variant sts uvs \<and> check_subsumes fresh_xx c (constraint_of constant ct uvs bef)) 
    (shows ''did not find '' +@+ 
     shows (case ct of Bound \<Rightarrow> ''bound'' | Strict \<Rightarrow> ''strict'' | NonStrict \<Rightarrow> ''non-strict'') +@+ 
     shows '' constraint for sequence '' +@+ shows_rules sts)) (init_conds st))"

fun shows_cc_aux :: "bool \<Rightarrow> ('f :: show, 'v :: show)cond_constraint \<Rightarrow> shows" where
  "shows_cc_aux b (CC_rewr s t) = (shows s +@+ shows '' = '' +@+ shows t)"
| "shows_cc_aux b (CC_cond stri (s,t)) = (shows s +@+ shows (if stri then '' > '' else '' >= '') +@+ shows t)"
| "shows_cc_aux b (CC_all x c) = (let s = shows ''ALL '' +@+ shows x +@+ shows ''. '' +@+ shows_cc_aux False c
    in (if b then shows ''('' +@+ s +@+ shows '')'' else s))"
| "shows_cc_aux b (CC_impl cs c2) = (''('' +#+ shows_list_gen (\<lambda> c. shows_cc_aux True c) ''True'' [] '' and '' [] cs +@+ shows '' => '' +@+ shows_cc_aux True c2 +@+ shows '')'')"

definition shows_cc :: "('f :: show, 'v :: show)cond_constraint \<Rightarrow> shows" where
  "shows_cc \<equiv> shows_cc_aux False"

datatype ('f,'v)cond_constraint_prf = 
  Final 
| Delete_Condition "('f,'v)cond_constraint" "('f,'v)cond_constraint_prf"
  (* new constraint and proof *)
| Different_Constructor "('f,'v)cond_constraint"
  (* the rewrite condition where there is a different constructor *)
| Same_Constructor "('f,'v)cond_constraint" "('f,'v)cond_constraint" "('f,'v)cond_constraint_prf"
  (* the rewrite condition where there is the same constructor + new total constraint + proof *)
| Variable_Equation 'v "('f,'v)term" "('f,'v)cond_constraint" "('f,'v)cond_constraint_prf"
  (* the subst. x / t + new total constraint + proof *)
| Funarg_Into_Var "('f,'v)cond_constraint" nat 'v "('f,'v)cond_constraint" "('f,'v)cond_constraint_prf"
  (* the rewrite condition + the index of the arg. + name of fresh var + new total constraint + proof *)
| Simplify_Condition "('f,'v)cond_constraint" "('f,'v)substL" "('f,'v)cond_constraint" "('f,'v)cond_constraint_prf"
  (* the quantified subformula + sigma + new total constraint + proof *)
| Induction "('f,'v)cond_constraint" "('f,'v)cond_constraint list" "(('f,'v)rule \<times> (('f,'v)term \<times> 'v list) list \<times> ('f,'v)cond_constraint \<times> ('f,'v)cond_constraint_prf) list"
  (* the rewrite condition + the conjunction that is used for induction + (renamed rule, subterm + ys - list, new constraint for rule, proof *)

lemma
  fixes P :: "('f, 'v) cond_constraint_prf \<Rightarrow> bool"
  assumes "P Final" and
  "\<And> c p. P p \<Longrightarrow> P (Delete_Condition c p)" and 
  "\<And> c. P(Different_Constructor c)" and 
  "\<And> c1 c2 p. P p \<Longrightarrow> P(Same_Constructor c1 c2 p)" and 
  "\<And> x t c p. P p \<Longrightarrow> P(Variable_Equation x t c p)" and 
  "\<And> c1 i x c2 p. P p \<Longrightarrow> P(Funarg_Into_Var c1 i x c2 p)" and 
  "\<And> c1 \<sigma> c2 p. P p \<Longrightarrow> P(Simplify_Condition c1 \<sigma> c2 p)" and 
  "\<And> c cs ics. (\<And> lr rs_ys c p. (lr,rs_ys,c,p) \<in> set ics \<Longrightarrow> P p) \<Longrightarrow> P(Induction c cs ics)"
  shows cond_constraint_prf_induct[case_names
    Final Delete_Condition
    Different_Constructor
    Same_Constructor
    Variable_Equation
    Funarg_Into_Var
    Simplify_Condition
    Induction, induct type: cond_constraint_prf]:
    "P c"
proof (induct rule: cond_constraint_prf.induct[THEN conjunct1, 
    of P "\<lambda> ics. \<forall> (_,_,_,p) \<in> set ics. P p" "\<lambda> (_,_,_,p). P p" "\<lambda> (_,_,p). P p" "\<lambda> (_,p). P p"])
  case 8
  show ?case 
    by (rule assms(8), insert 8, auto)
qed (insert assms, auto)

definition check_rys where 
  "check_rys D rt r rys \<equiv> case rys of (r',ys) \<Rightarrow> do {
               check (root r' = rt) (''root of '' +#+ shows r' +@+ '' is not '' +#+ shows (the rt));
               check (r \<unrhd> r') (shows r' +@+ '' is not a subterm of '' +#+ shows r);
               check_allm (\<lambda> f. check (\<not> D f) (''the defined symbol '' +#+ shows f +@+ '' occurs in the subterm '' +#+ shows r' +@+ shows '' of the rhs''))
                 (funas_args_term_list r');
               check_disjoint ys (vars_term_list r)
                 <+? (\<lambda> y. shows y +@+ '' occurs in '' +#+ shows r)
             }"


fun check_cc_prf :: "('f,string)rules \<Rightarrow> ('f \<times> nat \<Rightarrow> bool) \<Rightarrow> ('f \<times> nat)list \<Rightarrow> bool \<Rightarrow> (('f,string)rules \<Rightarrow> shows check) \<Rightarrow> ('f :: show,string)cond_constraint \<Rightarrow> ('f,string)cond_constraint_prf \<Rightarrow> ('f,string)c_constraint list result" where
  "check_cc_prf R D F m ortho cc Final = (case normalize_cc cc of 
    CC_impl [] (CC_cond stri st) \<Rightarrow> return [Unconditional_C stri st]
  | CC_impl [CC_cond stri uv] (CC_cond stri' st) \<Rightarrow> if stri = stri' then return [Conditional_C stri uv st] else
      error (shows ''problem in final constraint: different relations for finalizing '' +@+ shows_cc cc)
  | _ \<Rightarrow> error (shows ''problem in final constraint: it is neither a condition nor an implification of two conditions, but it is '' +@+ shows_nl +@+ shows_cc cc)
  )"
| "check_cc_prf R D F m ortho c (Delete_Condition d prf) = do {
     check (check_subsumes fresh_xx d c) 
       (''problem in delete conditions when switching from '' +#+ shows_nl +@+ shows_cc c +@+ '' to '' +#+ shows_nl +@+ shows_cc d);
     check_cc_prf R D F m ortho d prf
   }"
| "check_cc_prf R D F m ortho c (Different_Constructor d) = (case normalize_cc c of
     CC_impl cs c' \<Rightarrow> do {
       check (d \<in> set cs) (shows_cc d +@+ shows_nl +@+ ''is not a premise of '' +#+ shows_cc c);
       (case d of
         CC_rewr (Fun f ss) (Fun g ts) \<Rightarrow> do {
           check (\<not> D (f,length ss)) (shows f +@+ shows '' is defined'');
           check ((f,length ss) \<noteq> (g,length ts)) (''the root '' +#+ shows f +@+ shows '' is identical on both sides'');
           return []
         }
       | _ \<Rightarrow> error (shows_cc d +@+ shows '' is not a rewrite condition of the correct shape''))}
     <+? (\<lambda> s. ''problem in Different Constructor with rewrite condition '' +#+ shows_cc d +@+ shows_nl +@+ 
      ''on input contraint '' +#+ shows_nl +@+ shows_cc c +@+ shows_nl +@+ s))"
| "check_cc_prf R D F m ortho c (Same_Constructor d c' p) = (case normalize_cc c of
     CC_impl cs con \<Rightarrow> do { do {
       check (d \<in> set cs) (shows_cc d +@+ shows_nl +@+ ''is not a premise of '' +#+ shows_cc c);
       (case d of
         CC_rewr (Fun f ss) (Fun g ts) \<Rightarrow> do {
           check (\<not> D (f,length ss)) (shows f +@+ shows '' is defined'');
           check ((f,length ss) = (g,length ts)) (shows f +@+ '' and '' +#+ shows g +@+ shows '' are not identical'');
           let ds = cs @ map (\<lambda> (s,t). CC_rewr s t) (zip ss ts);
           let d' = CC_impl ds con;
           check (check_subsumes fresh_xx c' d') (''new constraint is '' +#+ shows_cc c' +@+ shows_nl +@+ ''but expected was '' +#+ shows_cc d')
         }
       | _ \<Rightarrow> error (shows_cc d +@+ shows '' is not a rewrite condition of the correct shape''))}
     <+? (\<lambda> s. shows ''problem in Same Constructor with rewrite condition '' +@+ shows_cc d +@+ shows_nl +@+ 
         '' when switching from '' +#+ shows_nl +@+ shows_cc c +@+ '' to '' +#+ shows_nl +@+ shows_cc c' +@+ shows_nl +@+ s);
     check_cc_prf R D F m ortho c' p})"
| "check_cc_prf R D F m ortho c (Variable_Equation x t d p) = (case normalize_cc c of
     CC_impl cs con \<Rightarrow> do { do {
       check (CC_rewr (Var x) t \<in> set cs \<or> CC_rewr t (Var x) \<in> set cs \<and> (\<forall> f \<in> set (funas_term_list t). \<not> D f)) 
         (''could not find '' +#+ shows_cc (CC_rewr (Var x) t) +@+ '' or reversed as a premise of '' +#+ shows_nl +@+ shows_cc c);
       let c' = fresh.cc_subst_apply fresh_xx c (Var(x := t), vars_term_list t);
       check (check_subsumes fresh_xx d c') (''new constraint is '' +#+ shows_cc d +@+ shows_nl +@+ ''but expected was '' +#+ shows_cc c')
       }
     <+? (\<lambda> s. shows ''problem in Variable Equation with substitution '' +@+ shows x +@+ shows ''/'' +@+ shows t +@+ '' to switch from'' 
         +#+ shows_nl +@+ shows_cc c +@+ shows_nl +@+ ''to'' +#+ shows_nl +@+ shows_cc d +@+ shows_nl +@+ s);
     check_cc_prf R D F m ortho d p})" 
| "check_cc_prf R D F m ortho c (Funarg_Into_Var c' i x d p) = (case normalize_cc c of
     CC_impl cs con \<Rightarrow> do { do {
       check (c' \<in> set cs) (shows_cc c' +@+ shows_nl +@+ ''is not a premise of '' +#+ shows_cc c);
       check (x \<notin> set (vars_cc_list c)) (''variable '' +#+ shows x +@+ shows '' is not fresh'');
       (case c' of
         CC_rewr (Fun f ss) q \<Rightarrow> do {
           check (i < length ss) (shows ''invalid position'');
           let (bef,p,aft) = (take i ss, ss ! i, drop (Suc i) ss);
           check_subseteq (funas_term_list p) F <+? (\<lambda> f. ''function symbol '' +#+ shows f +@+ shows '' is not allowed in argument'');
           let px = CC_rewr p (Var x);
           let fq = CC_rewr (Fun f (bef @ Var x # aft)) q;
           let ds = px # fq # cs; 
           let d' = CC_impl ds con;
           check (check_subsumes fresh_xx d d') (''new constraint is '' +#+ shows_cc d +@+ shows_nl +@+ ''but expected was '' +#+ shows_cc d')
         }
       | _ \<Rightarrow> error (shows_cc c' +@+ shows '' is not a rewrite condition of the correct shape''))}
     <+? (\<lambda> s. shows ''problem in introducing fresh variable '' +@+ shows x +@+ '' on '' +#+ shows (Suc i) +@+ ''-th argument of lhs of ''
       +#+ shows_cc c' +@+ '' to switch from '' +#+ 
       shows_cc c +@+ ''to'' +#+ shows_nl +@+ shows_cc d +@+ shows_nl +@+ s);
     check_cc_prf R D F m ortho d p})"
| "check_cc_prf R D F m ortho c (Simplify_Condition bc \<sigma> d p) = (case normalize_cc c of
     CC_impl cs \<psi> \<Rightarrow> do { do {
       check (bc \<in> set cs) (shows_cc bc +@+ shows_nl +@+ ''is not a premise of '' +#+ shows_cc c);
       let (ys,cc) = cc_unbound bc;
       let (\<phi>',\<psi>') = (case normalize_cc cc of CC_impl \<phi>' \<psi>' \<Rightarrow> (\<phi>',\<psi>'));
       let dom_ran = mk_subst_domain \<sigma>;
       check_subseteq (map fst dom_ran) ys 
         <+? (\<lambda> y. shows y +@+ shows '' is in the domain of sigma, but not a bound quantifier '');
       check_allm (\<lambda> fn. do {
           check (\<not> D fn) (''symbol '' +#+ shows fn +@+ shows '' is not allowed in range of sigma, as it is defined'');
           check (fn \<in> set F) (''symbol '' +#+ shows fn +@+ shows '' is not allowed in range of sigma, as it is not in F'')
         }) (concat (map (\<lambda> x_t. funas_term_list (snd x_t)) dom_ran));
       let vs = remdups (concat (map (\<lambda> x_t. vars_term_list (snd x_t)) dom_ran));
       let sigma = (\<lambda> c. fresh.cc_subst_apply fresh_xx c (mk_subst Var \<sigma>, vs));
       check_allm (\<lambda> c. check (\<exists> c' \<in> set cs. check_subsumes fresh_xx c' (sigma c))
         (shows_cc (sigma c) +@+ shows_nl +@+ shows ''is not contained as premise of the input implication'')) \<phi>';
       let d' = CC_impl (sigma \<psi>' # cs) \<psi>;
       check (check_subsumes fresh_xx d d') (''new constraint is '' +#+ shows_cc d +@+ shows_nl +@+ ''but expected was '' +#+ shows_cc d')
     }
     <+? (\<lambda> s. shows ''problem in Simplify Condition with substitution '' +@+ shows \<sigma> +@+ '' on IH'' +#+ shows_nl +@+ shows_cc bc
     +@+ shows_nl +@+ ''to switch from'' +#+ shows_nl +@+ shows_cc c +@+ shows_nl +@+ ''to'' +#+ shows_nl +@+ shows_cc d +@+ shows_nl +@+ s);
     check_cc_prf R D F m ortho d p })"
| "check_cc_prf R D F m ortho c (Induction d ccs ihs) = (case normalize_cc c of
     CC_impl cs c' \<Rightarrow> do { do {
       (if m then OK else ortho R) <+? (\<lambda> s. shows ''CR or minimality required'' +@+ shows_nl +@+ s);
       check_allm (\<lambda> cc. check (cc \<in> set cs) (shows_cc cc +@+ shows_nl +@+ ''is not a premise of '' +#+ shows_cc c)) (d # ccs);
       (case d of
         CC_rewr (Fun f xs) q \<Rightarrow> do {
           let cs = vars_cc_list (CC_impl (CC_rewr (Fun f xs) q # ccs) c');
           check ((\<forall> x \<in> set xs. is_Var x) \<and> distinct (map the_Var xs)) 
             (shows ''arguments of '' +@+ shows (Fun f xs) +@+ shows '' are not different variables'');
           let xs' = map the_Var xs;
           let rt = root (Fun f xs);
           check (mgu (Fun f xs) q = None) (shows ''lhs and rhs unify'');
           check_allm (\<lambda> lr. check (root (fst lr) = rt \<longrightarrow> (\<exists> lr' \<in> set (map (\<lambda> (r,_). r) ihs). lr =v lr' \<and> isOK(check_disjoint cs (vars_rule_list lr')))) (shows ''could not find variable renamed version of rule '' +@+ shows_rule lr)) R;
           check_allm (\<lambda> ((l,r),rys,cc,_). do {
             let cc' = fresh.cc_rule_constraint fresh_xx f (args l) r q xs' ccs c' rys;
             check_allm (check_rys D rt r) rys;
             check (check_subsumes fresh_xx cc cc') (''new constraint is '' +#+ shows_cc cc +@+ shows_nl +@+ ''but expected was '' +#+ shows_cc cc')
           } <+? (\<lambda> s. ''problem in constraint for rule '' +#+ shows_rule (l,r) +@+ shows_nl +@+ s) 
           ) ihs
         }
       | _ \<Rightarrow> error (shows_cc d +@+ shows '' is not a rewrite condition of the correct shape''))}
     <+? (\<lambda> s. shows ''problem in Induction rule with rewrite condition '' +@+ shows_cc d +@+ '' to switch from'' +#+ shows_nl 
       +@+ shows_cc c +@+ shows_nl +@+ ''to'' +#+ shows_nl +@+ 
       shows_list_gen (\<lambda>(_,_,c,_).shows_cc c +@+ shows_nl) [] [] [] [] ihs +@+ s);
   fcss <- Error_Monad.mapM (\<lambda> (_,_,c,p). check_cc_prf R D F m ortho c p) ihs;
   return (concat fcss)})
"

lemma check_cc_prf: assumes "fixed_trs_dep_order \<pi> R Q F S NS" and "check_cc_prf RR (defined R) FF m ortho cc prf = return fcss" and 
  "Ball (set fcss) (cc_satisfied F S NS)" and F: "F = set FF" and R: "R = set RR" 
  and ortho: "\<And> R. isOK(ortho R) \<Longrightarrow> CR (rstep (set R))"
  shows "fixed_trs_dep_order.cc_valid R Q F S NS nfs m cc"
proof -
  let ?chk = "check_cc_prf RR (defined R) FF m ortho"
  let ?UR = "(\<lambda> _ . R) :: ('a,string)term \<Rightarrow> ('a,string)trs"
  obtain c :: "'a" where True by auto
  interpret fixed_trs_dep_order \<pi> R Q F S NS nfs m c by fact
  interpret fixed_trs_dep_order_fresh \<pi> R Q F S NS nfs m fresh_xx c ?UR
    by (unfold_locales, rule fresh_string, auto)
  from assms have "?chk cc prf = return fcss" and "Ball (set fcss) (cc_satisfied F S NS)" by auto
  thus "cc_valid cc"
  proof (induct "prf" arbitrary: cc fcss)
    case (Final cc fcss)
    let ?n = "normalize_cc cc"
    obtain cs c where n: "?n = CC_impl cs c" unfolding normalize_cc_def by auto
    note check = Final(1)[simplified, unfolded n, simplified]
    have "cc_valid ?n"
    proof (cases cs)
      case Nil
      with check
      obtain stri st where c: "c = CC_cond stri st" using check by (cases c, auto)
      with Nil check have "fcss = [Unconditional_C stri st]" by auto
      with Final(2) have "cc_valid c" unfolding c cc_valid_def normal_F_subst_def by (cases st, cases stri, auto simp: rel_of_def)
      with Nil show ?thesis unfolding n by auto
    next
      case (Cons c1 ccs)
      with check 
      obtain stri st where c1: "c1 = CC_cond stri st" using check by (cases c1, auto)
      from Cons c1 check have ccs: "ccs = []" by (cases ccs, auto)
      note Cons = Cons[unfolded c1 ccs]
      from Cons check obtain stri' uv where c: "c = CC_cond stri' uv" using check by (cases c, auto)
      from check Cons c have c: "c = CC_cond stri uv" by (cases "stri = stri'", auto)
      from Cons c check have "fcss = [Conditional_C stri st uv]" by auto
      with Final(2) show "cc_valid ?n" unfolding n Cons c 
        by (cases st, cases uv, cases stri, auto simp: cc_valid_def normal_F_subst_def rel_of_def)
    qed
    thus ?case by (rule normalize_cc_valid)
  next
    case (Delete_Condition c p d fcss)
    from Delete_Condition(2)[simplified]
    have sub: "check_subsumes fresh_xx c d" 
      and rec: "?chk c p = return fcss" by auto
    from Delete_Condition(1)[OF rec Delete_Condition(3)] have "cc_valid c" .
    from cc_impliesE[OF check_subsumes[OF sub] this]
    show "cc_valid d" .
  next
    case (Different_Constructor c d fcc)
    obtain ds d' where d: "normalize_cc d = CC_impl ds d'" unfolding normalize_cc_def by auto
    note check = Different_Constructor(1)[simplified, unfolded d, simplified]
    from check obtain s t where c: "c = CC_rewr s t" by (cases c, auto)
    note check = check[unfolded c, simplified]
    from check obtain f ss where s: "s = Fun f ss" by (cases s, auto)
    note check = check[unfolded s, simplified]
    from check obtain g ts where t: "t = Fun g ts" by (cases t, auto)
    note check = check[unfolded t, simplified]
    show ?case
    proof (rule cc_simplify[OF constructor_different[OF d]])
      from check show "(f,length ss) \<noteq> (g,length ts)" by auto
    qed (insert check, auto)
  next
    case (Same_Constructor d c' p c fcc)
    obtain cs con where c: "normalize_cc c = CC_impl cs con" unfolding normalize_cc_def by auto
    note check = Same_Constructor(2)[simplified, unfolded c, simplified]
    from check obtain s t where d: "d = CC_rewr s t" by (cases d, auto)
    note check = check[unfolded d, simplified]
    from check obtain f ss where s: "s = Fun f ss" by (cases s, auto)
    note check = check[unfolded s, simplified]
    from check obtain ts where t: "t = Fun f ts" by (cases t, auto)
    note check = check[unfolded t Let_def, simplified]
    from check have "?chk c' p = return fcc" by auto
    from Same_Constructor(1)[OF this Same_Constructor(3)] have c': "cc_valid c'" by auto
    show ?case
    proof (rule cc_simplify[OF constructor_same[OF c]])
      show "c' \<longrightarrow>cc CC_impl (cs @ map (\<lambda>(s, t). CC_rewr s t) (zip ss ts)) con"
        by (rule check_subsumes, insert check, auto)
    qed (insert check c', auto)
  next
    case (Variable_Equation x t d p c fcc)
    obtain cs con where c: "normalize_cc c = CC_impl cs con" unfolding normalize_cc_def by auto
    note check = Variable_Equation(2)[simplified, unfolded c Let_def, simplified]
    from check have "?chk d p = return fcc" by auto
    from Variable_Equation(1)[OF this Variable_Equation(3)] have d: "Ball (set [d]) cc_valid" by auto    
    show ?case
    proof (rule cc_simplify[OF _ d])
      from check have "CC_rewr (Var x) t \<in> set cs \<or> CC_rewr t (Var x) \<in> set cs \<and> (\<forall>f\<in>funas_term t. \<not> defined R f)" by auto
      thus "cc_simplify c [d]"
      proof
        assume mem: "CC_rewr (Var x) t \<in> set cs"
        show ?thesis
          by (rule variable_left[OF c check_subsumes refl mem], insert check, auto)
      next
        assume "CC_rewr t (Var x) \<in> set cs \<and> (\<forall>f\<in>funas_term t. \<not> defined R f)"
        hence mem: "CC_rewr t (Var x) \<in> set cs" and d: "(\<forall>f\<in>funas_term t. \<not> defined R f)" by auto
        show ?thesis
          by (rule variable_right[OF c check_subsumes refl mem], insert check d, auto)
      qed
    qed
  next
    case (Funarg_Into_Var c' i x d p c fcc)
    obtain cs con where c: "normalize_cc c = CC_impl cs con" unfolding normalize_cc_def by auto
    note check = Funarg_Into_Var(2)[simplified, unfolded c, simplified]
    from check obtain s t where c': "c' = CC_rewr s t" by (cases c', auto)
    note check = check[unfolded c', simplified]
    from check obtain f ss where s: "s = Fun f ss" by (cases s, auto)
    note check = check[unfolded s Let_def, simplified]
    from check have "?chk d p = return fcc" by auto
    from Funarg_Into_Var(1)[OF this Funarg_Into_Var(3)] have d: "Ball (set [d]) cc_valid" by auto 
    from check have mem: "CC_rewr (Fun f ss) t \<in> set cs" and i: "i < length ss" by auto
    from mem id_take_nth_drop[OF i] have mem: "CC_rewr (Fun f (take i ss @ ss ! i # drop (Suc i) ss)) t \<in> set cs" by auto
    from check have x: "x \<notin> vars_cc c" by auto
    show ?case
      by (rule cc_simplify[OF fun_arg_into_var[OF c check_subsumes x mem] d], insert check F, auto)
  next
    case (Simplify_Condition bc \<sigma> d p c fcs)
    obtain cs con where c: "normalize_cc c = CC_impl cs con" unfolding normalize_cc_def by auto
    obtain ys cc where unb: "cc_unbound bc = (ys,cc)" by force
    obtain \<phi>' \<psi>' where cc: "normalize_cc cc = CC_impl \<phi>' \<psi>'" unfolding normalize_cc_def by auto
    note check = Simplify_Condition(2)[simplified, unfolded c, simplified, unfolded unb, simplified, unfolded cc Let_def, simplified]
    from check have "?chk d p = return fcs" by auto
    from Simplify_Condition(1)[OF this Simplify_Condition(3)] have valid: "Ball (set [d]) cc_valid" by simp
    from check cc_unbound_bound[OF unb] have mem: "cc_bound ys cc \<in> set cs" by auto
    let ?\<sigma> = "mk_subst Var \<sigma>"
    let ?vs = "remdups (concat (map (\<lambda>x_t. vars_term_list (snd x_t)) (mk_subst_domain \<sigma>)))"
    let ?\<sigma>vs = "(?\<sigma>, ?vs)"
    from check have subs: "check_subsumes fresh_xx d  (CC_impl (fresh.cc_subst_apply fresh_xx \<psi>' ?\<sigma>vs # cs) con)" by auto
    {
      fix c
      assume c: "c \<in> set \<phi>'"
      with check obtain d where d: "d \<in> set cs"
        and check: "check_subsumes fresh_xx d
               (fresh.cc_subst_apply fresh_xx c ?\<sigma>vs)" by auto
      from check_subsumes[OF check] d have "\<exists> d \<in> set cs. d \<longrightarrow>cc fresh.cc_subst_apply fresh_xx c ?\<sigma>vs" by blast
    } note subs2 = this
    show ?case
      by (rule cc_simplify[OF simplify_condition[OF c mem cc _ _ _ _ check_subsumes[OF subs]] valid],
        insert check subs2, auto simp: vars_subst_range_def mk_subst_domain F)
  next
    case (Induction d ccs ihs c fcss)
    obtain cs con where c: "normalize_cc c = CC_impl cs con" unfolding normalize_cc_def by auto
    note check = Induction(2)[simplified, unfolded c Let_def, simplified]
    from check obtain s q where d: "d = CC_rewr s q" by (cases d, auto)
    note check = check[unfolded d, simplified]
    from check obtain f xs where s: "s = Fun f xs" by (cases s, auto)
    note check = check[unfolded s, simplified]
    def xs' \<equiv> "map the_Var xs"
    note check = check[unfolded xs'_def[symmetric]]
    from check have dist: "distinct xs'" by auto
    from check have "\<And> x. x \<in> set xs \<Longrightarrow> is_Var x" by auto
    hence xs: "xs = map Var xs'" unfolding xs'_def map_map comp_def
      by (induct xs, auto)
    note check = check[unfolded xs, simplified]
    from check have mgu: "mgu (Fun f (map Var xs')) q = None" by auto
    from check have mem: "CC_rewr (Fun f (map Var xs')) q \<in> set cs" by auto
    from check have subset: "set ccs \<subseteq> set cs" by auto
    let ?c = "CC_impl (CC_rewr (Fun f (map Var xs')) q # ccs) con"
    let ?ccs = "map (\<lambda> (_,_,c,_). c) ihs"
    show ?case
    proof (rule cc_simplify[OF induction[OF c mgu mem subset _ dist]])
      fix ls r
      assume len: "length xs' = length ls"
        and mem: "(Fun f ls, r) \<in> R"
      with R check obtain lr' info where ihs: "(lr',info) \<in> set ihs" and
          eq: "(Fun f ls, r) =v lr'" and 
          disj: "vars_rule lr' \<inter> vars_cc ?c = {}" 
        by (auto simp: ac_simps)
      obtain l' r' where lr': "lr' = (l',r')" by force
      from eq_rule_mod_varsE[OF eq[unfolded lr']] obtain ls' where l': "l' = Fun f ls'" by (cases l', auto)
      obtain rys c p where info: "info = (rys,c,p)" by (cases info, auto)
      note id = l' info lr'
      note ihs = ihs[unfolded id]
      note eq = eq[unfolded id]
      note disj = disj[unfolded id]
      from ihs have c: "c \<in> set ?ccs" by force
      from check ihs have subs: "check_subsumes fresh_xx c
        (fresh.cc_rule_constraint fresh_xx f ls' r' q xs' ccs con rys)"
        by auto
      show "\<exists>d ls' r' rs_ys_list.
              (Fun f ls, r) =v (Fun f ls', r') \<and>
              vars_rule (Fun f ls', r') \<inter> vars_cc ?c = {} \<and>
              d \<in> set ?ccs \<and>
              d \<longrightarrow>cc fresh.cc_rule_constraint fresh_xx f ls' r' q xs' ccs con rs_ys_list \<and>
              (\<forall>(r, ys)\<in>set rs_ys_list.
                  root r = Some (f,length xs') \<and> r' \<unrhd> r \<and> vars_term r' \<inter> set ys = {} \<and> (\<forall>fn\<in>funas_args_term r. \<not> defined R fn))"
      proof (intro exI conjI, rule eq, rule disj, rule c, rule check_subsumes[OF subs], clarify)
        fix r ys
        assume "(r,ys) \<in> set rys"
        with check ihs have check: "isOK(check_rys (defined R) (Some (f, length xs')) r' (r,ys))" by force
        from this[unfolded check_rys_def]
        show "root r = Some (f, length xs') \<and> r' \<unrhd> r \<and> vars_term r' \<inter> set ys = {} \<and> (\<forall>fn\<in>funas_args_term r. \<not> defined R fn)" by auto
      qed
    next
      {
        fix c 
        assume c: "c \<in> set ?ccs"
        then obtain r i p where mem: "(r,i,c,p) \<in> set ihs" by force
        let ?rec = "Error_Monad.mapM (\<lambda>(_, _, c, p). ?chk c p) ihs"
        from check obtain fcssl where rec: "?rec = return fcssl"
          by (cases ?rec, auto)
        with check have fcss: "fcss = concat fcssl" by auto
        note res = mapM_Inr[OF rec]
        from res mem obtain fcs where chk: "?chk c p = return fcs" by (cases "?chk c p", auto)
        with res mem have "set fcs \<subseteq> set fcss" unfolding fcss by auto
        with Induction(3) have valid: "Ball (set fcs) (cc_satisfied F S NS)" by auto
        have "cc_valid c" 
          by (rule Induction(1)[OF mem chk valid])
      }
      thus "Ball (set ?ccs) cc_valid" by auto
    next
      from check have m: "m \<or> isOK(ortho RR)" by (cases m, auto)
      from m ortho[of RR] have m: "m \<or> CR(rstep (set RR))" by auto
      thus "m \<or> CR(rstep R)" 
        unfolding R by auto
    qed
  qed
qed

fun check_cc_prfs :: "('f,string)rules \<Rightarrow> ('f \<times> nat \<Rightarrow> bool) \<Rightarrow> ('f \<times> nat) list \<Rightarrow> bool \<Rightarrow> (('f,string)rules \<Rightarrow> shows check) \<Rightarrow> (('f :: show,string)cond_constraint \<times> 'a \<times> ('f,string)cond_constraint_prf) list \<Rightarrow> ('f,string)c_constraint list result" where
  "check_cc_prfs R D F m ortho [] = return []"
| "check_cc_prfs R D F m ortho ((c,_,prf) # cpfs) = (do {
    l1 <- check_cc_prf R D F m ortho c prf;
    l2 <- check_cc_prfs R D F m ortho cpfs;
    return (l1 @ l2)
   })"

lemma check_cc_prfs: assumes "fixed_trs_dep_order \<pi> R Q F S NS" and "check_cc_prfs RR (defined R) FF m ortho ccpfs = return fcss" and 
  "Ball (set fcss) (cc_satisfied F S NS)" and F: "F = set FF" and R: "R = set RR"
  and ortho: "\<And> R. isOK(ortho R) \<Longrightarrow> CR (rstep (set R))"
  shows "Ball (fst ` set ccpfs) (fixed_trs_dep_order.cc_valid R Q F S NS nfs m)"
proof -
  interpret fixed_trs_dep_order \<pi> R Q F S NS by fact
  let ?D = "defined R"
  from assms have "check_cc_prfs RR ?D FF m ortho ccpfs = return fcss" and "Ball (set fcss) (cc_satisfied F S NS)" by auto
  thus "Ball (fst ` set ccpfs) cc_valid"
  proof (induct ccpfs arbitrary: fcss)
    case Nil thus ?case by auto
  next
    case (Cons cpf ccpfs)
    obtain c a p where cpf: "cpf = (c,a,p)" by (cases cpf, auto)
    note check = Cons(2)[unfolded cpf, simplified]
    from check obtain l1 where cp: "check_cc_prf RR ?D FF m ortho c p = return l1" by (cases "check_cc_prf RR ?D FF m ortho c p", auto)
    from check cp obtain l2 where ccpfs: "check_cc_prfs RR ?D FF m ortho ccpfs = return l2" by (cases "check_cc_prfs RR ?D FF m ortho ccpfs", auto)
    from check cp ccpfs have fcss: "fcss = l1 @ l2" by auto
    note valid = Cons(3)[unfolded fcss]
    from Cons(1)[OF ccpfs] valid check_cc_prf[OF assms(1) cp _ _ _ ortho, of nfs] F R show ?case unfolding cpf by auto
  qed
qed


datatype ('f,'v)cond_red_pair_prf = Cond_Red_Pair_Prf 
  'f (* bound constant *)
  "(('f,'v)cond_constraint \<times> ('f,'v)rules \<times> ('f,'v)cond_constraint_prf) list" (* initial constraints with pair-sequence and proof *)
  nat (* how many pairs before are considered *)
  nat (* how many pairs afterwards are considered *)

definition
  conditional_general_reduction_pair_proc ::
    "('dpp, 'f :: {show,key}, string) dpp_ops \<Rightarrow> (('f \<times> nat) list \<Rightarrow> ('f,string) non_inf_order) \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)cond_red_pair_prf \<Rightarrow> bool \<Rightarrow> 'dpp \<Rightarrow> 'dpp list result"
where
  "conditional_general_reduction_pair_proc I grp Pstrict Pbound prof merge dpp \<equiv> case prof of
    Cond_Red_Pair_Prf c ccs bef aft \<Rightarrow> let 
    p = dpp_ops.pairs I dpp;
    r = dpp_ops.rules I dpp;
    F = remdups (funas_trs_list r @ funas_args_trs_list p @ concat (map funas_term_list (dpp_ops.Q I dpp)));
    rp = grp F in
  check_return (do {
     non_inf_order.valid rp;
     check (dpp_ops.wwf_rules I dpp) (shows ''require well-formedness of TRS'');
     let is_def = dpp_spec.is_defined I dpp;
     check_allm (\<lambda>(l, r). do {
        check_no_var l;
        check_no_var r;
        check_no_defined_root is_def r;
        check_subseteq (vars_term_list r) (vars_term_list l) 
          <+? (\<lambda> x. shows ''variable condition violated in pair '' +@+ shows_rule (l,r))         
      }) p;
     let ccs' = map (\<lambda> (c,uvs,prf). (c,uvs)) ccs;
     let check_present = check_constraint_present I dpp c p bef aft ccs';
     let (ps,pns) = dpp_ops.split_pairs I dpp Pstrict;
     let (pb,pnb) = dpp_ops.split_pairs I dpp Pbound;
     let pi = non_inf_order.af rp;
     let us = usable_rules_gen pi r p;
     let NS = check_allm (non_inf_order.ns rp);
     check_allm (check_present Strict) ps;
     check_allm (check_present Non_Strict) pns;
     check_allm (check_present Bound) pb;
     check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost required'');
     NS us 
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     let m = dpp_ops.minimal I dpp;
     (do {
        fcs \<leftarrow> check_cc_prfs r (dpp_spec.is_defined I dpp) F m check_weakly_orthogonal ccs
           <+? (\<lambda>s. ''problem when simplifying conditional constraints'' +#+ shows_nl +@+ s);
        check_allm (non_inf_order.cc rp) fcs
          <+? (\<lambda> s. ''problem when orienting final (conditional) constraints for pairs'' +#+ shows_nl +@+ s)
     })
   } <+? (\<lambda>s. ''could not apply the bounded increase processor with the following'' +#+ shows_nl +@+
     (non_inf_order.desc rp) +@+ shows_nl +@+ s))
    (if merge then [dpp_spec.delete_pairs I dpp (list_inter Pstrict Pbound)] else [dpp_spec.delete_pairs I dpp Pstrict, dpp_spec.delete_pairs I dpp Pbound])"

lemma conditional_general_reduction_pair_proc: fixes I :: "('a,'f :: {key,countable,show}, string)dpp_ops"
  assumes spec: "dpp_spec I"
    and grp: "generic_non_inf_order_impl grp"
    and proc: "conditional_general_reduction_pair_proc I (grp rp) Pstrict Pbound prof merge d = return dpps"
    and rec: "\<And>d'. d' \<in> set dpps \<Longrightarrow> finite_dpp (dpp_ops.dpp I d')"
  shows "finite_dpp (dpp_ops.dpp I d)"
proof -
  obtain c ccs bef aft where prof: "prof = Cond_Red_Pair_Prf c ccs bef aft" by (cases prof, force)
  note proc = proc[unfolded prof conditional_general_reduction_pair_proc_def cond_red_pair_prf.simps]
  interpret dpp_spec I by fact
  let ?RB = "rules d"
  let ?PB = "pairs d"
  let ?Q = "Q d"
  let ?M = "M d"
  let ?o = "check_weakly_orthogonal"
  def F \<equiv> "remdups (funas_trs_list ?RB @ funas_args_trs_list ?PB @ concat (map funas_term_list ?Q))"
  let ?CC = "non_inf_order.cc (grp rp F)"
  let ?NS = "non_inf_order.ns (grp rp F)"
  let ?pi = "non_inf_order.af (grp rp F)"
  obtain Ps Pns where split1: "dpp_ops.split_pairs I d Pstrict = (Ps,Pns)" by force
  obtain Pb Pub where split2: "dpp_ops.split_pairs I d Pbound = (Pb,Pub)" by force
  note proc = proc[unfolded Let_def split1 split2 F_def[symmetric], simplified]
  from proc have valid: "isOK (non_inf_order.valid (grp rp F))" by blast
  from proc have inn: "NF_terms (set (Q d)) \<subseteq> NF_trs (set (R d) \<union> set (Rw d))" by blast
  from proc have wwf: "wwf_qtrs (set ?Q) (set ?RB)" by simp
  from proc have dpps: "dpps = (if merge then [delete_pairs d (list_inter Pstrict Pbound)] else [delete_pairs d Pstrict, delete_pairs d Pbound])" by blast
  from proc have fun_ndef_vars: "\<And> s t. (s,t) \<in> set (P d) \<union> set (Pw d) \<Longrightarrow> 
    is_Fun s \<and> is_Fun t \<and> isOK (check_no_defined_root (defined (set (R d) \<union> set (Rw d))) t)
    \<and> vars_term t \<subseteq> vars_term s" by auto
  from proc obtain fcs where ccs: "check_cc_prfs (rules d) (defined (set (R d) \<union> set (Rw d))) F ?M ?o ccs = return fcs"  by (cases "check_cc_prfs (rules d) (defined (set (R d) \<union> set (Rw d))) F ?M ?o ccs", auto)
  note proc = proc[unfolded ccs, simplified]
  from proc have NS: "isOK(check_allm ?NS (usable_rules_gen (non_inf_order.af (grp rp F)) ?RB ?PB))" by auto
  from proc have CC: "isOK(check_allm ?CC fcs)" by auto
  let ?ccs' = "map (\<lambda> (c,uvs,prf). (c,uvs)) ccs"
  let ?present = "\<lambda> ct st. isOK(check_constraint_present I d c (pairs d) bef aft ?ccs' ct st)"
  from proc have presentS: "Ball (set Ps) (?present Strict)" by auto
  from proc have presentNS: "Ball (set Pns) (?present Non_Strict)" by auto
  from proc have presentB: "Ball (set Pb) (?present Bound)" by auto
  from generic_non_inf_order_impl.generate_non_inf_order[OF grp valid NS CC] obtain S NS where
    non_inf: "non_inf_order S NS (set F) ?pi" and 
    fcs: "Ball (set fcs) (cc_satisfied (set F) S NS)" and
    us: "set (usable_rules_gen ?pi ?RB ?PB) \<subseteq> NS" by force
  have inf: "infinite (UNIV :: string set)" by (rule infinite_UNIV_listI)
  interpret non_inf_order S NS "set F" ?pi by fact
  have fixed_trs: "fixed_trs_dep_order ?pi (set ?RB) (set ?Q) (set F) S NS" 
    by (unfold_locales, insert inn wwf F_def inf, auto)
  let ?UR = "\<lambda> _ . set ?RB"
  interpret fixed_trs_dep_order ?pi "set ?RB" "set ?Q" "set F" S NS "NFS d" ?M c by fact
  interpret fixed_trs_dep_order_fresh ?pi "set ?RB" "set ?Q" "set F" S NS "NFS d" ?M fresh_xx c ?UR
    by (unfold_locales, rule fresh_string, auto)
  note us = us[unfolded usable_rules_gen[OF refl]]
  have d: "dpp d = (NFS d, M d, set (P d), set (Pw d), set ?Q, set (R d), set (Rw d))" by simp
  {
    fix dd 
    assume dd: "dd \<in> set [delete_pairs d Pstrict, delete_pairs d Pbound]"     
    have "finite_dpp (dpp dd)" 
    proof (cases merge)
      case False
      with rec[unfolded dpps] dd show ?thesis by auto
    next
      case True
      with rec[unfolded dpps] have "finite_dpp (dpp (delete_pairs d (list_inter Pstrict Pbound)))" by simp
      note fin = finite_dpp_mono[OF this[unfolded delete_P_Pw_sound] _ _ refl subset_refl refl]
      let ?dpp = "\<lambda> PS. (NFS d, M d, set (P d) - set PS, set (Pw d) - set PS, set (Q d), set (R d), set (Rw d))"
      from dd have "dpp dd = ?dpp Pstrict \<or> dpp dd = ?dpp Pbound" unfolding delete_P_Pw_sound[symmetric] by auto
      then obtain PS where dd: "dpp dd = ?dpp PS" and mono: "set Pstrict \<inter> set Pbound \<subseteq> set PS" by auto
      show ?thesis unfolding dd by (rule fin, insert mono, auto)
    qed
  } note finite = this
  show ?thesis unfolding d
  proof (rule conditional_general_reduction_pair_proc)
    show "set (R d) \<union> set (Rw d) = set (rules d)" by simp
    show "gen_usable_rules_pairs (set (P d) \<union> set (Pw d)) \<subseteq> NS" using us by simp
    show "set (P d) \<union> set (Pw d) \<subseteq> set Ps \<union> set Pns" using split1 by auto
    show "funas_args_trs (set (P d) \<union> set (Pw d)) \<subseteq> set F" unfolding F_def by auto
    from ccs have ccs: "check_cc_prfs (rules d) (defined (set (rules d))) F ?M ?o ccs = Inr fcs" by simp
    show "Ball (fst ` set ccs) cc_valid"
      by (rule check_cc_prfs[OF fixed_trs ccs fcs refl refl check_weakly_orthogonal])
    let ?DS = "(NFS d, M d, set (P d) - set Ps, set (Pw d) - set Ps, set ?Q, set (R d), set (Rw d))"
    have "dpp (delete_pairs d Pstrict) = ?DS"
      unfolding delete_P_Pw_sound using split1 by auto
    with finite[of "delete_pairs d Pstrict"] 
    show "finite_dpp ?DS" by simp
    let ?DB = "(NFS d, M d, set (P d) - set Pb, set (Pw d) - set Pb, set ?Q, set (R d), set (Rw d))"
    have "dpp (delete_pairs d Pbound) = ?DB"
      unfolding delete_P_Pw_sound using split2 by auto
    with finite[of "delete_pairs d Pbound"] 
    show "finite_dpp ?DB" by simp
  next
    fix s t
    assume "(s,t) \<in> set (P d) \<union> set (Pw d)"
    from fun_ndef_vars[OF this] check_no_defined_root_sound[OF refl, of ?RB t]
    have "is_Fun s \<and> is_Fun t" and ndef: "\<not> defined (set ?RB) (the (root t))" and "vars_term t \<subseteq> vars_term s" by auto
    show "is_Fun s \<and> is_Fun t" by fact
    show "vars_term t \<subseteq> vars_term s" by fact
    show "\<not> defined (applicable_rules (set ?Q) (set ?RB)) (the (root t))" using ndef 
      unfolding defined_def applicable_rules_def by auto
  next
    {
      fix PP ct
      assume PP: "Ball (set PP) (?present ct)"
      have "Ball (set PP) (constraint_present bef aft (set (P d) \<union> set (Pw d)) ct (fst ` set ccs))" (is "Ball _ ?p")
      proof
        fix st
        assume "st \<in> set PP"
        with PP have st: "?present ct st" by auto
        note st = st[unfolded check_constraint_present_def Let_def]
        show "?p st" unfolding constraint_present_def
        proof (intro allI impI)
          fix sts
          assume sts: "sts \<in> initial_conditions bef aft (set (P d) \<union> set (Pw d)) st"
          have sts: "sts \<in> initial_conditions_gen (\<lambda>st uv. is_iedg_edge_dpp I d st (fst uv)) bef aft (set (P d) \<union> set (Pw d)) st"
            unfolding initial_conditions_def
            by (rule set_mp[OF initial_conditions_gen_mono sts[unfolded initial_conditions_def]], 
              insert is_iedg_edge_dpp_DG_sound[OF spec, where d = d], force)
          with st 
          obtain d uvs prof where mem: "(d,uvs,prof) \<in> set ccs" and  disj: "disjoint_variant sts uvs"
            and sub: "check_subsumes fresh_xx d (constraint_of c ct uvs bef)" by auto
          from mem check_subsumes[OF sub] disj
          show "\<exists>d uvs. disjoint_variant sts uvs \<and> d \<in> fst ` set ccs \<and> cc_implies d (constraint_of c ct uvs bef)" by force
        qed
      qed
    } note present = this
    show "Ball (set Ps) (constraint_present bef aft (set (P d) \<union> set (Pw d)) Strict (fst ` set ccs))"
      by (rule present[OF presentS])
    show "Ball (set Pns) (constraint_present bef aft (set (P d) \<union> set (Pw d)) Non_Strict (fst ` set ccs))"
      by (rule present[OF presentNS])
    show "Ball (set Pb) (constraint_present bef aft (set (P d) \<union> set (Pw d)) Bound (fst ` set ccs))"
      by (rule present[OF presentB])
  qed
qed

lemma conditional_general_reduction_pair_proc_len: 
  assumes "conditional_general_reduction_pair_proc I rp Pstrict Pbound prof merge d = return dpps"
  shows "length dpps = (if merge then 1 else 2)"
  using assms
  by (cases prof, cases merge, auto simp: conditional_general_reduction_pair_proc_def Let_def)

end
