(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Impl
imports
  Term
  Monadic_Check
  Show
begin

(* TODO: why cannot we use substpow.simps directly *)
primrec substpow_impl :: "('f, 'v) subst \<Rightarrow> nat \<Rightarrow> ('f, 'v) subst" where
  "substpow_impl \<sigma> 0  = Var"
| "substpow_impl \<sigma> (Suc n) = \<sigma> \<circ>s substpow_impl \<sigma> n"

lemma substpow_impl[simp]: "substpow_impl = (\<lambda> \<sigma> n . \<sigma> ^^ n)"
proof (rule ext, rule ext)
  fix \<sigma> n
  show "substpow_impl \<sigma> n = \<sigma> ^^ n"
    by (induct n, auto)
qed


subsection {*@{text "show"} Instance for Terms.*}

(*shows method for terms that is parametrized over shows methods for function
symbols and variables*)
fun shows_term' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('f, 'v) term \<Rightarrow> shows" where
  "shows_term' fun var (Var x) = var x"
| "shows_term' fun var (Fun f ts) = (
    fun f +@+ shows_list_gen id [] ''('' '', '' '')'' (map (shows_term' fun var) ts)
  )"

instantiation "term" :: ("show", "show") "show" begin
  abbreviation "shows_term t \<equiv> shows_term' shows shows t"
  definition "shows_prec (d::nat) t = shows_term t"
  definition "shows_list_term (is::('a::show, 'b::show) term list) = shows_list_aux shows is"

  lemma assoc_term:
    "shows_prec d (t::('a::show, 'b::show) term) r @ s = shows_prec d t (r @ s)"
  proof (induct t arbitrary: r s)
    case (Var x) show ?case unfolding shows_prec_term_def by simp
  next
    case (Fun f ts)
    hence "\<And>r s. \<forall>t\<in>set ts. id (shows_prec d t) r @ s = id (shows_prec d t) (r @ s)" by auto
    hence elt: "\<And>r s. \<forall>t\<in>set (map shows_term ts). id t r @ s = id t (r @ s)"
      unfolding shows_prec_term_def by auto
    show ?case
      unfolding shows_prec_term_def
      unfolding shows_term'.simps o_def
      unfolding assoc
      unfolding shows_list_gen_assoc[OF elt,of "[]" "''(''" "'', ''" "'')''" r s]
      using Fun by simp
  qed
  lemma assoc_term_elt:
    "\<forall>(t::('a::show, 'b::show) term)\<in>set ts. shows t r @ s = shows t (r @ s)"
    using assoc_term by auto
 instance by
    (intro_classes,rule assoc_term,unfold shows_list_term_def)
    (rule shows_list_aux_assoc[OF assoc_term_elt])
end


subsection {* @{text "shows"} Instance for Contexts *}

fun shows_ctxt' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> shows" where
  "shows_ctxt' fun var (Hole) = shows_string ''[]''"
| "shows_ctxt' fun var (More f ss1 D ss2) = (
    fun f +@+ ''('' +#+
    shows_list_gen (shows_term' fun var) [] [] '', '' '', '' ss1 +@+
    shows_ctxt' fun var D +@+
    shows_list_gen (shows_term' fun var) '')'' '', '' '', '' '')'' ss2
  )"

instantiation ctxt :: ("show", "show") "show" begin
  abbreviation "shows_ctxt C \<equiv> shows_ctxt' shows shows C"
  definition "shows_prec (d::nat) C = shows_ctxt C"
  definition "shows_list_ctxt (cs::('a::show,'b::show) ctxt list) \<equiv> shows_list_aux shows cs"

  lemma assoc_ctxt:
    "shows_prec d (c::('a::show, 'b::show) ctxt) r @ s = shows_prec d c (r @ s)"
    by (induct c arbitrary: r s)
       (auto simp: shows_prec_ctxt_def shows_prec_term_def[symmetric, of _ 0])

  lemma assoc_ctxt_elt:
    "\<forall>(c::('a::show, 'b::show) ctxt)\<in>set cs. shows_prec d c r @ s = shows_prec d c (r @ s)"
  using assoc_ctxt by auto

  instance by
    (intro_classes,rule assoc_ctxt,unfold shows_list_ctxt_def)
    (rule shows_list_aux_assoc[OF assoc_ctxt_elt])
end

fun
  poss_list :: "('f, 'v) term \<Rightarrow> pos list"
where
  "poss_list (Var x) = [\<epsilon>]" |
  "poss_list (Fun f ss) = (\<epsilon> # concat (map (\<lambda> (i,ps).
    map ( op <# i) ps) (zip [0 ..< length ss] (map poss_list ss))))"

lemma poss_list_sound[simp]: "set (poss_list s) = poss s" 
proof (induct s, simp)
  case (Fun f ss)
  let ?z = "zip [0..<length ss] (map poss_list ss)"
  have len: "length ?z = length ss" by auto
  have "(\<Union>a\<in>set ?z.
         set (split (\<lambda>i. map (op <# i)) a)) = 
       {i <# p |i p. i < length ss \<and> p \<in> poss (ss ! i)}" (is "?l = ?r")
  proof (rule set_eqI)
    fix ip
    show "(ip \<in> ?l) = (ip \<in> ?r)"
    proof
      assume "ip \<in> ?l" 
      from this obtain ipI where 
	z: "ipI \<in> set ?z" and 
	ip: "ip \<in> set (split (\<lambda> i. map (op <# i)) ipI)" 
	by auto     
      from z obtain i where i: "i < length ?z" and zi: "?z ! i = ipI"
        by (force simp: set_conv_nth)      
      with ip Fun show "ip \<in> ?r" by auto
    next
      assume "ip \<in> ?r"
      from this obtain i p where i: "i < length ss" and "p \<in> poss (ss ! i)"
	and ip: "ip = i <# p" by auto
      with Fun have p: "p \<in> set (poss_list (ss ! i))" and iz: "i < length ?z" by auto
      from i have id: "?z ! i = (i, poss_list (ss ! i))" (is "_ = ?ipI") by auto
      from iz have  "?z ! i \<in> set ?z" by (rule nth_mem)
      with id have inZ: "?ipI \<in> set ?z" by auto
      from p have "i <# p \<in> set (split (\<lambda> i. map (op <# i)) ?ipI)" by auto
      with inZ ip show "ip \<in> ?l" by force
    qed
  qed
  with Fun show ?case by simp
qed

declare poss_list.simps[simp del]

fun in_poss :: "pos \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "in_poss \<epsilon> _ = True"
      | "in_poss (PCons i p) (Fun f ts) = (i < length ts \<and> in_poss p (ts ! i))"
      | "in_poss (PCons i p) (Var _) = False"

lemma in_poss[simp]: "in_poss p t = (p \<in> poss t)"
  by (induct rule: in_poss.induct, auto)

declare in_poss.simps[simp del]

lemma [code]: "p \<in> poss t = in_poss p t" by simp

text {*Resolve problems with recurisve dependencies in generated code.*}
lemma [code]:
  shows "HOL.eq (Var x) (Var y) \<longleftrightarrow> HOL.eq x y"
  and "HOL.eq (Fun f ts) (Fun g us) \<longleftrightarrow> HOL.eq f g \<and> list_all2 HOL.eq ts us"
  by (simp_all add: list_all2_eq [symmetric])

subsection {* Useful abstractions *}

text {* Given that we perform the same operations on terms in order to get
a list of the variables and a list of the functions, we define functions
that run through the term and perform these actions. *}

fun contains_var_term :: "'v \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "contains_var_term x (Var y) = (x = y)"
| "contains_var_term x (Fun _ ts) = Bex (set ts) (contains_var_term x)"

lemma contains_var_term_sound[simp]:
  "contains_var_term x t \<longleftrightarrow> x \<in> vars_term t"
  by (induct t) auto

(*use efficient implementation for code-generation*)
lemma [code_unfold]: "x \<in> vars_term t = contains_var_term x t" by simp

fun vars_term_list' :: "('f, 'v) term \<Rightarrow> 'v list" where
  "vars_term_list' (Var x) = [x]"
| "vars_term_list' (Fun _ ts) = concat (map vars_term_list' ts)"

lemma vars_term_list'_sound[simp]: "set (vars_term_list' t) = vars_term t"
  by (induct t) simp_all

lemma unary_vars_term_list': assumes t: "funas_term t \<subseteq> F"
  and unary: "\<And> f n. (f,n) \<in> F \<Longrightarrow> n \<le> 1"
  shows "vars_term_list' t = [] \<or> (\<exists> x \<in> vars_term t. vars_term_list' t = [x])"
proof -
  from t show ?thesis
  proof (induct t)
    case Var thus ?case by auto
  next
    case (Fun f ss)
    show ?case
    proof (cases ss)
      case Nil
      thus ?thesis by auto
    next
      case (Cons t ts)
      let ?n = "length ss"
      from Fun(2) have "(f,?n) \<in> F" by auto
      from unary[OF this] have n: "?n \<le> Suc 0" by auto
      with Cons have "?n = Suc 0" by simp
      with Cons have ss: "ss = [t]" by (cases ts, auto)
      note IH = Fun(1)[unfolded ss, simplified]
      from ss have id1: "vars_term_list' (Fun f ss) = vars_term_list' t" by simp
      from ss have id2: "vars_term (Fun f ss) = vars_term t" by simp
      from Fun(2) ss have mem: "funas_term t \<subseteq> F" by auto
      show ?thesis unfolding id1 id2 using IH[OF refl mem] by simp
    qed
  qed
qed

declare vars_term_list'.simps[simp del]


definition vars_term_list :: "('f, 'v) term \<Rightarrow> 'v list" where
  "vars_term_list = remdups \<circ> vars_term_list'"

lemma vars_term_list_sound[simp]:
  "set (vars_term_list t) = vars_term t"
  by (simp add: vars_term_list_def)

text {*The list of function symbols in a term (without removing duplicates).*}

fun funs_term_list' :: "('f, 'v) term \<Rightarrow> 'f list" where
  "funs_term_list' (Var _) = []"
| "funs_term_list' (Fun f ts) = f # concat (map funs_term_list' ts)"

lemma funs_term_list'_sound[simp]:
  "set (funs_term_list' t) = funs_term t"
  by (induct t) simp_all
declare funs_term_list'.simps[simp del]

definition funs_term_list :: "('f, 'v) term \<Rightarrow> 'f list" where
  "funs_term_list \<equiv> remdups \<circ> funs_term_list'"

lemma funs_term_list_sound[simp]:
  "set (funs_term_list t) = funs_term t"
  by (simp add: funs_term_list_def)

abbreviation funs_terms_list where "funs_terms_list Q \<equiv> remdups (concat (map funs_term_list Q))"


text {*The list of function symbol and arity pairs in a term
(without removing duplicates).*}

fun funas_term_list' :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_term_list' (Var _) = []"
| "funas_term_list' (Fun f ts) = (f, length ts) # concat (map funas_term_list' ts)"

lemma funas_term_list'_sound[simp]:
  "set (funas_term_list' t) = funas_term t"
  by (induct t) simp_all
declare funas_term_list'.simps[simp del]

definition funas_term_list :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_term_list \<equiv> remdups \<circ> funas_term_list'"

lemma funas_term_list_sound[simp]:
  "set (funas_term_list t) = funas_term t"
  by (simp add: funas_term_list_def)

definition funas_args_term_list' :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_args_term_list' t = concat (map funas_term_list' (args t))"

lemma funas_args_term_list'_sound[simp]:
  "set (funas_args_term_list' t) = funas_args_term t"
  unfolding funas_args_term_def funas_args_term_list'_def by simp

definition funas_args_term_list :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_args_term_list = remdups \<circ> funas_args_term_list'"

lemma funas_args_term_list_sound[simp]:
  "set (funas_args_term_list t) = funas_args_term t"
  by (simp add: funas_args_term_list_def)

lemma vars_term_list'_map_term: "vars_term_list' (map_term fg t) = vars_term_list' t"
proof (induct t)
  case (Var x) thus ?case by (simp add: vars_term_list'.simps)
next
  case (Fun f ss)
  show ?case by (simp add: vars_term_list'.simps o_def, insert Fun, induct ss, auto)
qed

lemma funs_term_list'_map_term:
  "funs_term_list' (map_term fg t) = map fg (funs_term_list' t)"
proof (induct t)
  case (Var x) show ?case by (simp add: funs_term_list'.simps)
next
  case (Fun f ts)
  show ?case 
    by (simp add: funs_term_list'.simps, insert Fun, induct ts, auto)
qed

lemma linear_vars_term_list': assumes "linear_term t"
  shows "length (filter (op = x) (vars_term_list' t)) \<le> 1"
  using assms
proof (induct t)
  case (Var y)
  show ?case by (auto simp: vars_term_list'.simps)
next
  case (Fun f ss)
  show ?case 
  proof (rule ccontr)
    assume "\<not> ?thesis"
    from this[unfolded vars_term_list'.simps]
    have len: "2 \<le> length (filter (op = x) (concat (map vars_term_list' ss)))" (is "_ \<le> length ?xs") by auto
    from len obtain y ys where xs: "?xs = y # ys" by (cases ?xs, auto)
    from len[unfolded xs] obtain z zs where ys: "ys = z # zs" by (cases ys, auto)
    from xs[unfolded ys] have "{y,z} \<subseteq> set ?xs" by auto
    from this[unfolded set_filter] have "y = x" and "z = x" by auto
    from xs[unfolded ys this] have xs: "?xs = x # x # zs" by auto
    {
      fix s
      assume s: "s \<in> set ss"
      with Fun(2)[unfolded linear_term.simps]
      have "linear_term s" by auto
      note Fun(1)[OF s this]
    }
    from this Fun(2)[unfolded linear_term.simps] xs
    show False
    proof (induct ss)
      case Nil thus ?case by simp
    next
      case (Cons s ss) note oCons = this
      from Cons(3) have part: "is_partition (map vars_term ss) \<and> (\<forall> s \<in> set ss. linear_term s)" and lin: "linear_term s"  using is_partition_Cons by auto
      from Cons(1)[OF _ part] Cons(2) 
      have ind: "filter (op = x) (concat (map vars_term_list' ss)) \<noteq> x # x # zs" by auto
      show ?case 
      proof (cases "filter (op = x) (vars_term_list' s)")
        case Nil
        with Cons(4) ind show False by auto
      next
        case (Cons y ys)
        let ?s = "filter (op = x) (vars_term_list' s)"
        let ?ss = "filter (op = x) (concat (map vars_term_list' ss))"
        from Cons oCons(4) have "?s = x # ys" by auto
        with oCons(2)[of s] have sx: "?s = [x]"
          by auto
        with oCons(4) have ssx: "?ss = x # zs" by auto
        from sx have "x \<in> set ?s" by auto
        from this[unfolded set_filter vars_term_list'_sound]
        have sx: "x \<in> vars_term s" by auto 
        from ssx have "x \<in> set ?ss" by auto
        from this[unfolded set_filter vars_term_list'_sound]
        obtain t where tx: "x \<in> vars_term t" and t: "t \<in> set ss" by auto
        from t[unfolded set_conv_nth] obtain i where i: "i < length ss" and
          t: "t = ss ! i" by auto
        from oCons(3)[THEN conjunct1, unfolded is_partition_def, THEN spec[of _ "Suc i"], THEN mp, THEN spec[of _ 0]] i tx[unfolded t] sx
        show False by auto
      qed
    qed  
  qed
qed


lemma vars_term_list_map_term:
  "vars_term_list (map_term fg t) = vars_term_list t"
  unfolding vars_term_list_def vars_term_list'_map_term o_def ..

fun supteq_list :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list" where
  "supteq_list (Var x) = [Var x]"
| "supteq_list (Fun f ts) = Fun f ts # concat (map supteq_list ts)"

fun supt_list :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list" where
  "supt_list (Var x) = []"
| "supt_list (Fun f ts) = concat (map supteq_list ts)"

lemma supteq_list[simp]: "set (supteq_list t) = {s. t \<unrhd> s}"
proof (rule set_eqI, simp)
  fix s
  show "s \<in> set(supteq_list t) = (t \<unrhd> s)"
  proof (induct t, simp add: supteq_var_imp_eq)
    case (Fun f ss)
    show ?case
    proof (cases "Fun f ss = s", (auto)[1])
      case False
      show ?thesis
      proof
        assume "Fun f ss \<unrhd> s"
        with False have sup: "Fun f ss \<rhd> s" using supteq_supt_conv by auto
        obtain C where "C \<noteq> \<box>" and "Fun f ss = C\<langle>s\<rangle>" using sup by auto
        then obtain b D a where "Fun f ss = Fun f (b @ D\<langle>s\<rangle> # a)" by (cases C, auto)
        hence D: "D\<langle>s\<rangle> \<in> set ss" by auto
        with Fun[OF D] ctxt_imp_supteq[of D s] obtain t where "t \<in> set ss" and "s \<in> set (supteq_list t)" by auto
        thus "s \<in> set (supteq_list (Fun f ss))" by auto
      next
        assume "s \<in> set (supteq_list (Fun f ss))"
        with False obtain t where t: "t \<in> set ss" and "s \<in> set (supteq_list t)" by auto
        with Fun[OF t] have "t \<unrhd> s" by auto
        with t show "Fun f ss \<unrhd> s" by auto
      qed
    qed
  qed
qed

lemma supt_list_sound[simp]: "set (supt_list t) = {s. t \<rhd> s}"
  by (cases t) auto

fun supt_impl :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "supt_impl (Var x) t = False"
| "supt_impl (Fun f ss) t = (t \<in> set ss \<or> Bex (set ss) (\<lambda>s. supt_impl s t))"

lemma supt_impl[code_unfold]: "s \<rhd> t \<longleftrightarrow> supt_impl s t"
proof
  assume "s \<rhd> t" thus "supt_impl s t"
  proof (induct s)
    case (Var x) thus ?case by auto
  next
    case (Fun f ss) thus ?case
    proof (cases "t \<in> set ss")
      case True thus ?thesis by (simp)
    next
      case False
      assume "\<And>s. \<lbrakk>s \<in> set ss; s \<rhd> t\<rbrakk> \<Longrightarrow> supt_impl s t"
        and "Fun f ss \<rhd> t" and "t \<notin> set ss"
      moreover from `Fun f ss \<rhd> t` obtain s where "s \<in> set ss" and "s \<rhd> t"
        by (cases rule: supt.cases) (simp_all add: `t \<notin> set ss`)
      ultimately have "supt_impl s t" by simp
      with `s \<in> set ss` show ?thesis by auto
    qed
  qed
next
  assume "supt_impl s t"
  thus "s \<rhd> t"
  proof (induct s)
    case (Var x) thus ?case by simp
  next
    case (Fun f ss)
    hence "t \<in> set ss \<or> (\<exists>s\<in>set ss. supt_impl s t)" by simp
    thus ?case
    proof
      assume "t \<in> set ss" thus ?case by (auto intro: supt.intros)
    next
      assume "\<exists>s\<in>set ss. supt_impl s t"
      then obtain s where "s \<in> set ss" and "supt_impl s t" by auto
      with Fun have "s \<rhd> t" by auto
      with `s \<in> set ss` show ?thesis by (auto intro: supt.intros)
    qed
  qed
qed

lemma supteq_impl[code_unfold]: "s \<unrhd> t \<longleftrightarrow> s = t \<or> supt_impl s t"
  unfolding supteq_supt_set_conv
  by (auto simp: supt_impl)

fun
  linear_term_impl :: "'v list \<Rightarrow> ('f, 'v) term \<Rightarrow> ('v list) option"
where
  "linear_term_impl xs (Var x) = (if x \<in> set xs then None else Some (x # xs))" |
  "linear_term_impl xs (Fun _ []) = Some xs" |
  "linear_term_impl xs (Fun f (t # ts)) = (case linear_term_impl xs t of 
    None \<Rightarrow> None
  | Some ys \<Rightarrow> linear_term_impl ys (Fun f ts))"

lemma [code]: "linear_term t = (case linear_term_impl [] t of None \<Rightarrow> False | Some _ \<Rightarrow> True)" 
proof -
  {
    fix xs ys
    have "(linear_term_impl xs t = None \<longrightarrow> (set xs \<inter> vars_term t \<noteq> {} \<or> \<not> linear_term t)) \<and> 
      (linear_term_impl xs t = Some ys \<longrightarrow> (set ys = set xs \<union> vars_term t) \<and> set xs \<inter> vars_term t = {} \<and> linear_term t)"
    proof (induct arbitrary: ys rule: linear_term_impl.induct, force, simp add: is_partition_Nil)
      case (3 xs f t ts zs)
      show ?case 
      proof (cases "linear_term_impl xs t")
        case None
        with 3 show ?thesis by auto
      next
        case (Some ys)
        note some = this
        with 3 have rec1: "set ys = set xs \<union> vars_term t \<and> set xs \<inter> vars_term t = {} \<and> linear_term t" by auto
        show ?thesis 
        proof (cases "linear_term_impl ys (Fun f ts)")
          case None 
          with rec1 Some have res: "linear_term_impl xs (Fun f (t # ts)) = None" by simp
          from None 3(2) Some have rec2: "set ys \<inter> vars_term (Fun f ts) \<noteq> {} \<or> \<not> linear_term (Fun f ts)" by simp
          hence "set xs \<inter> vars_term (Fun f (t # ts)) \<noteq> {} \<or> \<not> linear_term (Fun f (t # ts))" 
          proof 
            assume "set ys \<inter> vars_term (Fun f ts) \<noteq> {}" 
            then obtain x where x1: "x \<in> set ys" and x2: "x \<in> vars_term (Fun f ts)" by auto
            show ?thesis 
            proof (cases "x \<in> set xs")
              case True
              with x2 show ?thesis by auto
            next
              case False
              with x1 rec1 have "x \<in> vars_term t" by auto
              with x2 have "\<not> linear_term (Fun f (t # ts))" by (auto simp: is_partition_Cons)
              thus ?thesis ..
            qed
          next
            assume "\<not> linear_term (Fun f ts)" hence "\<not> linear_term (Fun f (t # ts))" by (auto simp: is_partition_Cons)
            thus ?thesis ..
          qed            
          with res show ?thesis by auto
        next
          case (Some us)
          with some have res: "linear_term_impl xs (Fun f (t # ts)) = Some us" by auto          
          {
            assume us: "us = zs"
            from Some[simplified us] 3(2) some 
            have rec2: "set zs = set ys \<union> vars_term (Fun f ts) \<and> set ys \<inter> vars_term (Fun f ts) = {} \<and> linear_term (Fun f ts)" by auto
            from rec1 rec2  
            have part1: "set zs = set xs \<union> vars_term (Fun f (t # ts)) \<and> set xs \<inter> vars_term (Fun f (t # ts)) = {}" (is ?part1) by auto
            from rec1 rec2 have "vars_term t \<inter> vars_term (Fun f ts) = {}" and "linear_term t" and "linear_term (Fun f ts)" by auto
            hence "linear_term (Fun f (t # ts))" (is ?part2) by (auto simp: is_partition_Cons) 
            with part1 have "?part1 \<and> ?part2" ..
          }
          with res show ?thesis by auto
        qed
      qed
    qed
  } note main = this
  from main[of "[]"] show ?thesis by (cases "linear_term_impl [] t", auto)
qed

definition check_no_var :: "('f::show, 'v::show) term \<Rightarrow> shows check" where
  "check_no_var t \<equiv> check (is_Fun t) (''variable found'' +#+ shows_nl)"

lemma check_no_var_sound[simp]:
  "isOK (check_no_var t) \<longleftrightarrow> is_Fun t"
  using assms unfolding check_no_var_def by simp


definition
  check_supt :: "('f::show, 'v::show) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_supt s t \<equiv> check (s \<rhd> t) (shows t +@+ '' is not a proper subterm of '' +#+ shows s)"

definition
  check_supteq :: "('f::show, 'v::show) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_supteq s t \<equiv> check (s \<unrhd> t) (shows t +@+ '' is not a subterm of '' +#+ shows s)"

lemma isOK_check_supt[simp]:
  "isOK (check_supt s t) = ((s, t) \<in> op \<rhd>)"
unfolding check_supt_def using supt_def by auto

lemma isOK_check_supteq[simp]:
  "isOK (check_supteq s t) = ((s, t) \<in> op \<unrhd>)"
unfolding check_supteq_def using supteq_def by auto

subsection {*Additional Functions on Terms*}

fun with_arity :: "('f, 'v) term \<Rightarrow> ('f \<times> nat, 'v) term" where
  "with_arity (Var x) = Var x"
| "with_arity (Fun f ts) = Fun (f, length ts) (map with_arity ts)"

fun add_vars_term_list :: "('f, 'v) term \<Rightarrow> 'v list \<Rightarrow> 'v list" where
  "add_vars_term_list (Var x) xs = x # xs"
| "add_vars_term_list (Fun _ ts) xs = foldr add_vars_term_list ts xs"

fun add_funs_term_list :: "('f, 'v) term \<Rightarrow> 'f list \<Rightarrow> 'f list" where
  "add_funs_term_list (Var _) fs = fs"
| "add_funs_term_list (Fun f ts) fs = f # foldr add_funs_term_list ts fs"

lemma foldr_append_map_append [simp]:
  "foldr (op @) (map f xs) ys @ zs = foldr (\<lambda>x. op @ (f x)) xs (ys @ zs)"
  by (induct xs) simp_all

lemma add_vars_term_list_vars_term_list'_conv:
  "add_vars_term_list t xs = vars_term_list' t @ xs"
  by (induct t arbitrary: xs)
    (simp_all cong: foldr_cong add: vars_term_list'.simps concat_conv_foldr)

lemma add_funs_term_list_funs_term_list'_conv:
  "add_funs_term_list t fs = funs_term_list' t @ fs"
  by (induct t arbitrary: fs)
    (simp_all cong: foldr_cong add: funs_term_list'.simps concat_conv_foldr)

lemma [code]:
  "vars_term_list' t = add_vars_term_list t []"
  "funs_term_list' t = add_funs_term_list t []"
  by (simp_all add: add_vars_term_list_vars_term_list'_conv
                    add_funs_term_list_funs_term_list'_conv)

(*FIXME: define funas via with_arity and add_funs*)

lemma with_arity_term_fold[code]:
  "with_arity = Term.fold Var (\<lambda>f ts. Fun (f, length ts) ts)"
proof
  fix t show "with_arity t = Term.fold Var (\<lambda>f ts. Fun (f, length ts) ts) t"
    by (induct t) simp_all
qed


fun flatten_term_enum :: "('f list,'v)term \<Rightarrow> ('f,'v)term list"
where "flatten_term_enum (Var x) = [Var x]"
   |  "flatten_term_enum (Fun fs ts) = (
          let lts = map flatten_term_enum ts;
              ss = concat_lists lts
            in concat (map (\<lambda> f. map (Fun f) ss) fs))"

lemma flatten_term_enum: 
  shows "set (flatten_term_enum t) = {u. instance_term u (map_term set t)}"
proof (induct t)
  case (Var x)
  show ?case (is "_ = ?R")
  proof -
    {
      fix t 
      assume "t \<in> ?R"
      hence "t = Var x" by (cases t, auto)
    } 
    thus ?thesis by auto
  qed
next
  case (Fun fs ts)
  show ?case (is "?L = ?R")
  proof -
    {
      fix i
      assume "i < length ts"
      hence "ts ! i \<in> set ts" by auto
      note Fun[OF this]
    } note ind = this
    have idL: "?L = {Fun g ss | g ss. g \<in> set fs  \<and> length ss = length ts \<and> (\<forall>i<length ts. ss ! i \<in> set (flatten_term_enum (ts ! i)))}" (is "_ = ?M1") by auto  
    let ?R1 = "{Fun f ss | f ss. f \<in> set fs \<and> length ss = length ts \<and> (\<forall> i<length ss. instance_term (ss ! i) (map_term set (ts ! i)))}"
    {
      fix u
      assume "u \<in> ?R"
      hence "u \<in> ?R1" by (cases u, auto)
    }
    hence idR: "?R = ?R1" by auto
    show ?case unfolding idL idR using ind by auto
  qed
qed

subsection {* Creating substitutions from lists *}

type_synonym ('f, 'v, 'w) gsubstL = "('v \<times> ('f, 'w) term) list"
type_synonym ('f, 'v) substL = "('f, 'v, 'v) gsubstL"

definition mk_subst :: "('v \<Rightarrow> ('f, 'w) term) \<Rightarrow> ('f, 'v, 'w) gsubstL \<Rightarrow> ('f, 'v, 'w) gsubst" where
  "mk_subst d xts \<equiv>
    (\<lambda>x. case map_of xts x of
      None \<Rightarrow> d x
    | Some t \<Rightarrow> t)"

lemma mk_subst_not_mem: assumes x: "x \<notin> set xs"
  shows "mk_subst f (zip xs ts) x = f x"
proof -
  have "map_of (zip xs ts) x = None"
    unfolding map_of_eq_None_iff set_zip using x[unfolded set_conv_nth] by auto
  thus ?thesis unfolding mk_subst_def by auto
qed

lemma mk_subst_distinct: assumes dist: "distinct xs"
  and len: "length ls = length xs"
  and i: "i < length xs"
  shows "mk_subst f (zip xs ls) (xs ! i) = ls ! i"
proof -
  from map_of_zip_is_None[OF len[symmetric], of "xs ! i"] i
  obtain l where map: "map_of (zip xs ls) (xs ! i) = Some l" by auto
  from map_of_SomeD[OF map] have "(xs ! i, l) \<in> set (zip xs ls)" by auto
  from this[unfolded set_conv_nth length_zip len] obtain j where
    j: "j < length xs" and id: "(xs ! i, l) = zip xs ls ! j" by auto
  from arg_cong[OF id, of fst] j len have "xs ! i = xs ! j" by auto
  with nth_eq_iff_index_eq[OF dist i j] have "i = j" by auto
  with id j len have l: "l = ls ! i" by auto
  with map show ?thesis  unfolding mk_subst_def by auto
qed

lemma mk_subst_Nil [simp]:
  "mk_subst d [] = d"
  by (simp add: mk_subst_def)

(* TODO: perhaps implement more efficiently *)
definition mk_subst_domain :: "('f, 'v) substL \<Rightarrow> ('v \<times> ('f, 'v) term) list" where
  "mk_subst_domain \<sigma> \<equiv>
    let \<tau> = mk_subst Var \<sigma> in
    remdups (filter (\<lambda>(x, t). Var x \<noteq> t) (map (\<lambda>(x, t). (x, \<tau> x)) \<sigma>))"

lemma mk_subst_domain:
  "set (mk_subst_domain \<sigma>) = (\<lambda> x. (x, mk_subst Var \<sigma> x)) ` subst_domain (mk_subst Var \<sigma>)"
  (is "?I = ?R")
proof -
  have "?I \<subseteq> ?R" unfolding mk_subst_domain_def Let_def subst_domain_def by auto
  moreover 
  {
    fix xt
    assume mem: "xt \<in> ?R"    
    obtain x t where xt: "xt = (x, t)" by force
    from mem [unfolded xt]
      have x: "x \<in> subst_domain (mk_subst Var \<sigma>)" and t: "t = mk_subst Var \<sigma> x" by auto
    hence "mk_subst Var \<sigma> x \<noteq> Var x" unfolding subst_domain_def by simp
    with t have l: "map_of \<sigma> x = Some t" and tx: "t \<noteq> Var x" 
      unfolding mk_subst_def by (cases "map_of \<sigma> x", auto)
    from map_of_SomeD[OF l] l t tx have "(x,t) \<in> ?I" unfolding mk_subst_domain_def Let_def
      by force
    hence "xt \<in> ?I" unfolding xt .
  }
  ultimately show ?thesis by blast
qed

lemma finite_mk_subst: "finite (subst_domain (mk_subst Var \<sigma>))"
proof -
  have "subst_domain (mk_subst Var \<sigma>) = fst ` set (mk_subst_domain \<sigma>)"
    unfolding mk_subst_domain Let_def by force
  moreover have "finite ..."
    using finite_set by auto
  ultimately show ?thesis by simp
qed

definition subst_eq :: "('f, 'v) substL \<Rightarrow> ('f, 'v) substL \<Rightarrow> bool" where
  "subst_eq \<sigma> \<tau> = (let \<sigma>' = mk_subst_domain \<sigma>; \<tau>' = mk_subst_domain \<tau> in set \<sigma>' = set \<tau>')"  

lemma subst_eq [simp]:
  "subst_eq \<sigma> \<tau> = (mk_subst Var \<sigma> = mk_subst Var \<tau>)"
proof -
  let ?\<sigma> = "mk_subst Var \<sigma>"
  let ?\<tau> = "mk_subst Var \<tau>"
  {
    assume id: "((\<lambda> x. (x, ?\<sigma> x)) ` subst_domain ?\<sigma>) =  ((\<lambda> x. (x, ?\<tau> x)) ` subst_domain ?\<tau>)" (is "?l = ?r")
    from arg_cong[OF id, of "op ` fst"] have idd: "subst_domain ?\<sigma> = subst_domain ?\<tau>" by force
    have "?\<sigma> = ?\<tau>" 
    proof (rule ext)
      fix x
      show "?\<sigma> x = ?\<tau> x"
      proof (cases "x \<in> subst_domain ?\<sigma>")
        case False
        thus ?thesis using idd unfolding subst_domain_def by auto
      next
        case True
        with idd have x: "(x,?\<sigma> x) \<in> ?l" "(x,?\<tau> x) \<in> ?r" by auto
        with id have x: "(x,?\<tau> x) \<in> ?l" "(x,?\<sigma> x) \<in> ?l" by auto
        thus ?thesis by auto
      qed
    qed
  }
  thus ?thesis 
    unfolding subst_eq_def Let_def
    unfolding mk_subst_domain by auto
qed

definition vars_subst_range_impl :: "('f, 'v) substL \<Rightarrow> 'v list" where 
  "vars_subst_range_impl \<sigma> \<equiv> let \<sigma>' = mk_subst_domain \<sigma> in  
      concat (map (vars_term_list o snd) \<sigma>')"

definition vars_subst_impl :: "('f, 'v) substL \<Rightarrow> 'v list" where 
  "vars_subst_impl \<sigma> \<equiv> let \<sigma>' = mk_subst_domain \<sigma> in map fst \<sigma>' @ 
      concat (map (vars_term_list o snd) \<sigma>')"

lemma vars_subst_impl [simp]:
  "set (vars_subst_impl \<sigma>) = vars_subst (mk_subst Var \<sigma>)"
  unfolding vars_subst_def vars_subst_impl_def Let_def 
  by (auto simp: mk_subst_domain, force)

lemma vars_subst_range_impl [simp]:
  "set (vars_subst_range_impl \<sigma>) = vars_subst_range (mk_subst Var \<sigma>)"
  unfolding vars_subst_range_def vars_subst_range_impl_def Let_def 
  by (auto simp: mk_subst_domain)

lemma mk_subst_one [simp]: "mk_subst Var [(x, t)] = {x / t}"
  unfolding mk_subst_def subst_def by auto

lemma fst_image [simp]: "fst ` (\<lambda> x. (x,g x)) ` a = a" by force

definition
  subst_compose_impl :: "('f, 'v) substL \<Rightarrow> ('f, 'v) substL \<Rightarrow> ('f, 'v) substL"
where
  "subst_compose_impl \<sigma> \<tau> \<equiv> 
  let
    \<sigma>' = mk_subst_domain \<sigma>; 
    \<tau>' = mk_subst_domain \<tau>;
    d\<sigma> = map fst \<sigma>'
  in map (\<lambda> (x, t). (x, t \<cdot> mk_subst Var \<tau>')) \<sigma>' @ filter (\<lambda> (x, t). x \<notin> set d\<sigma>) \<tau>'"

lemma mk_subst_mk_subst_domain [simp]:
  "mk_subst Var (mk_subst_domain \<sigma>) = mk_subst Var \<sigma>"
proof (intro ext)
  fix x
  {
    assume x: "x \<notin> subst_domain (mk_subst Var \<sigma>)"
    hence \<sigma>: "mk_subst Var \<sigma> x = Var x" unfolding subst_domain_def by auto
    from x have "x \<notin> fst ` set (mk_subst_domain \<sigma>)" unfolding mk_subst_domain by auto
    hence look: "map_of (mk_subst_domain \<sigma>) x = None" by (cases "map_of (mk_subst_domain \<sigma>) x", insert map_of_SomeD[of "mk_subst_domain \<sigma>" x], force+)
    hence "mk_subst Var (mk_subst_domain \<sigma>) x = mk_subst Var \<sigma> x" unfolding \<sigma>
      unfolding mk_subst_def by auto
  } note ndom = this
  {
    assume "x \<in> subst_domain (mk_subst Var \<sigma>)"
    hence "x \<in> fst ` set (mk_subst_domain \<sigma>)" unfolding mk_subst_domain by auto
    then obtain t where look: "map_of (mk_subst_domain \<sigma>) x = Some t" by (cases "map_of (mk_subst_domain \<sigma>) x", (force simp: map_of_eq_None_iff)+)
    from map_of_SomeD[OF look, unfolded mk_subst_domain] have t: "t = mk_subst Var \<sigma> x" by auto
    from look t have res: "mk_subst Var (mk_subst_domain \<sigma>) x = mk_subst Var \<sigma> x" unfolding mk_subst_def by auto
  } note dom = this
  from ndom dom
  show "mk_subst Var (mk_subst_domain \<sigma>) x = mk_subst Var \<sigma> x" by auto
qed

lemma subst_compose_impl [simp]:
  "mk_subst Var (subst_compose_impl \<sigma> \<tau>) = mk_subst Var \<sigma> \<circ>s mk_subst Var \<tau>" (is "?l = ?r")
proof (rule ext)
  fix x
  let ?\<sigma> = "mk_subst Var \<sigma>"
  let ?\<tau> = "mk_subst Var \<tau>"
  let ?s = "map (\<lambda> (x, t). (x, t \<cdot> mk_subst Var (mk_subst_domain \<tau>))) (mk_subst_domain \<sigma>)"
  let ?t = "[(x,t) \<leftarrow> mk_subst_domain \<tau>. x \<notin> set (map fst (mk_subst_domain \<sigma>))]"
  note d = subst_compose_impl_def[unfolded Let_def]
  show "?l x = ?r x"
  proof (cases "x \<in> subst_domain (mk_subst Var \<sigma>)")
    case True
    hence "?\<sigma> x \<noteq> Var x" unfolding subst_domain_def by auto
    then obtain t where look: "map_of \<sigma> x = Some t" and \<sigma>: "?\<sigma> x = t"
      unfolding mk_subst_def by (cases "map_of \<sigma> x", auto)
    from \<sigma> have r: "?r x = t \<cdot> ?\<tau>" unfolding subst_compose_def by simp
    from True have "x \<in> subst_domain (mk_subst Var (mk_subst_domain \<sigma>))" 
      by simp
    from \<sigma> True  have mem: "(x,t \<cdot> ?\<tau>) \<in> set ?s" by (auto simp: mk_subst_domain)
    with map_of_eq_None_iff[of "?s" x]
      obtain u where look2: "map_of ?s x = Some u"
      by (cases "map_of ?s x", force+)
    from map_of_SomeD[OF this] \<sigma> have u: "u = t \<cdot> ?\<tau>" 
      by (auto simp: mk_subst_domain)
    note look2 = map_of_append_Some[OF look2, of ?t]
    have l: "?l x = t \<cdot> ?\<tau>" unfolding d mk_subst_def[of Var "?s @ ?t"] look2 u
      by simp
    from l r show ?thesis by simp
  next
    case False
    hence \<sigma>: "?\<sigma> x = Var x" unfolding subst_domain_def by auto
    from \<sigma> have r: "?r x = ?\<tau> x" unfolding subst_compose_def by simp
    from False have "x \<notin> subst_domain (mk_subst Var (mk_subst_domain \<sigma>))" 
      by simp
    from False  have mem: "\<And> y. (x,y) \<notin> set ?s" by (auto simp: mk_subst_domain)
    with map_of_SomeD[of ?s x] have look2: "map_of ?s x = None"
      by (cases "map_of ?s x", auto)
    note look2 = map_of_append_None[OF look2, of ?t]
    have l: "?l x = (case map_of ?t x of None \<Rightarrow> Var x | Some t \<Rightarrow> t)" unfolding d mk_subst_def[of Var "?s @ ?t"] look2 by simp
    also have "... = ?\<tau> x"
    proof (cases "x \<in> subst_domain ?\<tau>")
      case True
      hence "?\<tau> x \<noteq> Var x" unfolding subst_domain_def by auto
      then obtain t where look: "map_of \<tau> x = Some t" and \<tau>: "?\<tau> x = t"
        unfolding mk_subst_def by (cases "map_of \<tau> x", auto)
      from True have "x \<in> subst_domain (mk_subst Var (mk_subst_domain \<tau>))" 
        by simp
      from \<tau> True  have mem: "(x,?\<tau> x) \<in> set ?t" using False by (auto simp: mk_subst_domain)
      with map_of_eq_None_iff[of ?t x] obtain u where look2: "map_of ?t x = Some u"
        by (cases "map_of ?t x", force+)
      from map_of_SomeD[OF this] \<tau> have u: "u = ?\<tau> x" 
        by (auto simp: mk_subst_domain)
      show ?thesis using look2 u by simp
    next
      case False
      hence \<tau>: "?\<tau> x = Var x" unfolding subst_domain_def by auto
      from False have "x \<notin> subst_domain (mk_subst Var (mk_subst_domain \<tau>))" 
        by simp
      from False have mem: "\<And> y. (x,y) \<notin> set ?t" by (auto simp: mk_subst_domain)
      with map_of_SomeD[of ?t x] have look2: "map_of ?t x = None"
        by (cases "map_of ?t x", auto)
      show ?thesis unfolding \<tau> look2 by simp
    qed
    finally show ?thesis unfolding r by simp
  qed
qed

fun subst_power_impl :: "('f, 'v) substL \<Rightarrow> nat \<Rightarrow> ('f, 'v) substL" where
  "subst_power_impl \<sigma> 0 = []"
| "subst_power_impl \<sigma> (Suc n) = subst_compose_impl \<sigma> (subst_power_impl \<sigma> n)"

lemma subst_power_impl [simp]:
  "mk_subst Var (subst_power_impl \<sigma> n) = (mk_subst Var \<sigma>)^^n"
  by (induct n, auto)
        
definition commutes_impl :: "('f, 'v) substL \<Rightarrow> ('f, 'v) substL \<Rightarrow> bool" where
  "commutes_impl \<sigma> \<mu> \<equiv> subst_eq (subst_compose_impl \<sigma> \<mu>) (subst_compose_impl \<mu> \<sigma>)"

lemma commutes_impl [simp]:
  "commutes_impl \<sigma> \<mu> = ((mk_subst Var \<sigma> \<circ>s mk_subst Var \<mu>) = (mk_subst Var \<mu> \<circ>s mk_subst Var \<sigma>))"
  unfolding commutes_impl_def by simp

definition
  subst_compose'_impl :: "('f, 'v) substL \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) substL"
where
  "subst_compose'_impl \<sigma> \<rho> \<equiv> map (\<lambda> (x, s). (x, s \<cdot> \<rho>)) (mk_subst_domain \<sigma>)"

lemma subst_compose'_impl [simp]:
  "mk_subst Var (subst_compose'_impl \<sigma> \<rho>) = subst_compose' (mk_subst Var \<sigma>) \<rho>" (is "?l = ?r")
proof (rule ext)
  fix x
  note d = subst_compose'_def subst_compose'_impl_def 
  let ?\<sigma> = "mk_subst Var \<sigma>"
  let ?s = "subst_compose'_impl \<sigma> \<rho>"
  show "?l x = ?r x"
  proof (cases "x \<in> subst_domain (mk_subst Var \<sigma>)")
    case True
    hence r: "?r x = ?\<sigma> x \<cdot> \<rho>" unfolding d by simp
    from True have "(x, ?\<sigma> x) \<in> set (mk_subst_domain \<sigma>)" unfolding mk_subst_domain by auto
    hence "(x, ?\<sigma> x \<cdot> \<rho>) \<in> set ?s" unfolding d by auto
    with map_of_eq_None_iff[of ?s x] obtain u where look: "map_of ?s x = Some u"
      by (cases "map_of ?s x", force+)
    from map_of_SomeD[OF this] have u: "u = ?\<sigma> x \<cdot> \<rho>" unfolding d using mk_subst_domain[of \<sigma>] by auto
    hence l: "?l x = ?\<sigma> x \<cdot> \<rho>" using look u unfolding mk_subst_def by auto
    from l r show ?thesis by simp
  next
    case False
    hence r: "?r x = Var x" unfolding d by simp
    from False have "\<And> y. (x,y) \<notin> set ?s" unfolding d 
      by (auto simp: mk_subst_domain)
    with map_of_SomeD[of ?s x] have look: "map_of ?s x = None" 
      by (cases "map_of ?s x", auto)
    hence l: "?l x = Var x" unfolding mk_subst_def by simp
    from l r show ?thesis by simp
  qed
qed

definition
  subst_replace_impl :: "('f, 'v) substL \<Rightarrow> 'v \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) substL"
where
  "subst_replace_impl \<sigma> x t \<equiv> (x, t) # filter (\<lambda> (y, t). y \<noteq> x) \<sigma>"

lemma subst_replace_impl [simp]:
  "mk_subst Var (subst_replace_impl \<sigma> x t) = (\<lambda> y. if x = y then t else mk_subst Var \<sigma> y)" (is "?l = ?r")
proof (rule ext)
  fix y
  note d = subst_replace_impl_def
  show "?l y = ?r y"
  proof (cases "y = x")
    case True
    thus ?thesis unfolding d mk_subst_def by auto
  next
    case False
    let ?\<sigma> = "mk_subst Var \<sigma>"
    from False have r: "?r y = ?\<sigma> y" by auto
    from False have l: "?l y =  mk_subst Var ([(y, t)\<leftarrow>\<sigma> . y \<noteq> x]) y" unfolding mk_subst_def d
      by simp
    also have "... = ?\<sigma> y" unfolding mk_subst_def  
      using map_of_filter[of "\<lambda> y. y \<noteq> x" y \<sigma>, OF False] by simp
    finally show ?thesis using r by simp
  qed          
qed

lemma mk_subst_domain_distinct: "distinct (map fst (mk_subst_domain \<sigma>))" 
  unfolding mk_subst_domain_def Let_def distinct_map set_remdups set_filter
  by (rule conjI[OF distinct_remdups], unfold inj_on_def, auto)

(* TODO: perhaps generalize all mk_subst things to arbitrary functions 
  which are represented as finitely many key-value pairs *)
definition var_renaming_impl :: "('f,'v) substL \<Rightarrow> bool" where
  "var_renaming_impl \<sigma> \<equiv>
    let \<sigma>' = map snd (mk_subst_domain \<sigma>) in
    (\<forall> t \<in> set \<sigma>'. is_Var t) \<and> distinct \<sigma>'"

lemma var_renaming_impl [simp]:
  "var_renaming_impl \<sigma> = var_renaming (mk_subst Var \<sigma>)" (is "?l = ?r")
proof -
  let ?\<sigma> = "mk_subst Var \<sigma>"
  let ?d = "mk_subst_domain \<sigma>"
  let ?m = "map snd ?d"
  let ?k = "map fst ?d"
  have "?l = ((\<forall> t \<in> set ?m. is_Var t) \<and> distinct ?m)" unfolding var_renaming_impl_def Let_def by auto
  also have "(\<forall> t \<in> set ?m. is_Var t) = (\<forall> x. is_Var (?\<sigma> x))" 
    by (force simp: mk_subst_domain subst_domain_def)
  also have "distinct ?m = inj_on ?\<sigma> (subst_domain ?\<sigma>)" 
  proof
    assume inj: "inj_on ?\<sigma> (subst_domain ?\<sigma>)"
    show "distinct ?m" unfolding distinct_conv_nth 
    proof (intro allI impI)
      fix i j
      assume i: "i < length ?m" and j: "j < length ?m" and ij: "i \<noteq> j"
      obtain x t where di: "?d ! i = (x,t)" by (cases "?d ! i", auto)
      obtain y s where dj: "?d ! j = (y,s)" by (cases "?d ! j", auto)
      from di i have mi: "?m ! i = t" and ki: "?k ! i = x" by auto
      from dj j have mj: "?m ! j = s" and kj: "?k ! j = y" by auto
      from di i have xt: "(x,t) \<in> set ?d" unfolding set_conv_nth by force
      from dj j have ys: "(y,s) \<in> set ?d" unfolding set_conv_nth by force
      from xt ys have d: "x \<in> subst_domain ?\<sigma>" "y \<in> subst_domain ?\<sigma>" unfolding mk_subst_domain by auto
      have dist: "distinct ?k" by (rule mk_subst_domain_distinct)
      from ij i j have xy: "x \<noteq> y" unfolding ki[symmetric] kj[symmetric] 
        using dist[unfolded distinct_conv_nth] by auto
      from xt ys have m: "?\<sigma> x = t" "?\<sigma> y = s" unfolding mk_subst_domain by auto
      from inj[unfolded inj_on_def, rule_format, OF d]
      show "?m ! i \<noteq> ?m ! j" unfolding m mi mj using xy by auto
    qed
  next      
    assume dist: "distinct ?m"
    show "inj_on ?\<sigma> (subst_domain ?\<sigma>)" unfolding inj_on_def
    proof (intro ballI impI)
      fix x y
      assume x: "x \<in> subst_domain ?\<sigma>" and y: "y \<in> subst_domain ?\<sigma>" 
        and id: "?\<sigma> x = ?\<sigma> y" 
      from x y have x: "(x,?\<sigma> x) \<in> set ?d" and y: "(y,?\<sigma> y) \<in> set ?d"
        unfolding mk_subst_domain by auto
      from x obtain i where di: "?d ! i = (x,?\<sigma> x)" and i: "i < length ?d" unfolding set_conv_nth by auto
      from y obtain j where dj: "?d ! j = (y,?\<sigma> y)" and j: "j < length ?d" unfolding set_conv_nth by auto
      from di i have mi: "?m ! i = ?\<sigma> x" by simp
      from dj j have mj: "?m ! j = ?\<sigma> x" unfolding id by simp
      from mi mj have id: "?m ! i = ?m ! j" by simp
      from dist[unfolded distinct_conv_nth] i j id have id: "i = j" by auto
      with di dj
      show "x = y" by auto
    qed
  qed
  finally
  show ?thesis unfolding var_renaming_def by simp
qed

definition inverse_var_renaming_impl :: "('f, 'v) substL \<Rightarrow> ('f, 'v) substL" where
  "inverse_var_renaming_impl \<sigma> \<equiv>
    let \<sigma>' = mk_subst_domain \<sigma> in
    map (\<lambda> (x, y). (the_Var y, Var x)) \<sigma>'"

lemma inverse_var_renaming_impl [simp]:
  fixes \<sigma> :: "('f, 'v) substL"
  assumes var: "var_renaming (mk_subst Var \<sigma>)"
  shows "mk_subst Var (inverse_var_renaming_impl \<sigma>) = inverse_var_renaming (mk_subst Var \<sigma>)" (is "?l = ?r")
proof (rule ext)
  fix x
  let ?\<sigma> = "mk_subst Var \<sigma>"
  let ?\<sigma>' = "mk_subst_domain \<sigma>"
  let ?m = "map (\<lambda> (x, y). (the_Var y, Var x :: ('f, 'v) term)) ?\<sigma>'"
  let ?ran = "subst_range ?\<sigma>"
  note d = inverse_var_renaming_impl_def inverse_var_renaming_def
  {
    fix t
    assume "(x,t) \<in> set ?m" 
    then obtain u z where id: "(x,t) = (the_Var u,Var z)" and mem: "(z,u) \<in> set ?\<sigma>'" by auto
    from var[unfolded var_renaming_def] mem obtain zz where u: "u = Var zz" 
      unfolding mk_subst_domain by auto
    from id[unfolded u] have id: "zz = x" "t = Var z" by auto
    with mem u have "(z,Var x) \<in> set ?\<sigma>'" by auto
    hence "?\<sigma> z = Var x" "z \<in> subst_domain ?\<sigma>" unfolding mk_subst_domain by auto
    with id have "\<exists> z. t = Var z \<and> ?\<sigma> z = Var x \<and> z \<in> subst_domain ?\<sigma>" by auto
  } note one = this
  have "?l x = mk_subst Var ?m x" unfolding d by simp
  also have "... = ?r x" 
  proof (cases "Var x \<in> ?ran")
    case False
    {
      fix t
      assume "(x,t) \<in> set ?m"
      from one[OF this] obtain z where t: "t = Var z" and z: "?\<sigma> z = Var x" 
        and dom: "z \<in> subst_domain ?\<sigma>" by auto
      from z dom False have False by force
    }
    from this[OF map_of_SomeD[of ?m x]] have look: "map_of ?m x = None" 
      by (cases "map_of ?m x", auto)    
    hence "mk_subst Var ?m x = Var x" unfolding mk_subst_def by auto
    also have "... = ?r x" using False unfolding d by simp
    finally show ?thesis .
  next
    case True
    then obtain y where y: "y \<in> subst_domain ?\<sigma>" and x: "?\<sigma> y = Var x" by auto
    hence "(y,Var x) \<in> set ?\<sigma>'" unfolding mk_subst_domain by auto
    hence "(x,Var y) \<in> set ?m" by force
    then obtain u where look: "map_of ?m x = Some u" using map_of_eq_None_iff[of ?m x]
      by (cases "map_of ?m x", force+)
    from map_of_SomeD[OF this] have xu: "(x,u) \<in> set ?m" by auto
    from one[OF this] obtain z where u: "u = Var z" and z: "?\<sigma> z = Var x" and dom: "z \<in> subst_domain ?\<sigma>" by auto
    have "mk_subst Var ?m x = Var z" unfolding mk_subst_def look u by simp
    also have "... = ?r x" using var_renaming_inverse_domain[OF var dom] z by auto
    finally show ?thesis .
  qed
  finally show "?l x = ?r x" .
qed  

definition
  mk_subst_case :: "'v list \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) substL \<Rightarrow> ('f, 'v) substL"
where
  "mk_subst_case xs \<sigma> \<tau> = subst_compose_impl (map (\<lambda> x. (x, \<sigma> x)) xs) \<tau>"

lemma mk_subst_case [simp]:
  "mk_subst Var (mk_subst_case xs \<sigma> \<tau>) =
    (\<lambda> x. if x \<in> set xs then \<sigma> x \<cdot> mk_subst Var \<tau> else mk_subst Var \<tau> x)" 
proof -
  let ?m = "map (\<lambda> x. (x, \<sigma> x)) xs"
  have id: "mk_subst Var ?m = (\<lambda> x. if x \<in> set xs then \<sigma> x else Var x)" (is "?l = ?r")
  proof (rule ext)
    fix x
    show "?l x = ?r x"
    proof (cases "x \<in> set xs")
      case True
      hence "(x,\<sigma> x) \<in> set ?m" by auto
      with map_of_eq_None_iff[of ?m x] obtain u where look: "map_of ?m x = Some u" by auto
      from map_of_SomeD[OF look] have u: "u = \<sigma> x" by auto
      show ?thesis unfolding mk_subst_def look u using True by auto
    next
      case False
      with map_of_SomeD[of ?m x]
        have look: "map_of ?m x = None" by (cases "map_of ?m x", auto)
      show ?thesis unfolding mk_subst_def look using False by auto
    qed
  qed
  show ?thesis unfolding mk_subst_case_def subst_compose_impl id 
    unfolding subst_compose_def by auto
qed


subsection {*Code Generator Setup*}

text{*The target language is assumed to be Haskell. Hence the following
code equations try to increase lazyness as far as possible.*}
lemma [code]:
  "add_vars_term_list = Term.fold (op #) (const (flip (foldr (op \<circ>)) id))"
    (is "_ = ?fun")
proof (intro ext)
  fix t xs show "add_vars_term_list t xs = ?fun t xs"
  proof (induct t arbitrary: xs)
    case (Var x) show ?case by simp
  next
    case (Fun f ts xs)
    from foldr_cong[of xs xs ts ts add_vars_term_list, OF _ _ this]
      have IH: "foldr add_vars_term_list ts xs = foldr ?fun ts xs" by blast
    show ?case
      by (simp add: IH foldr_map const_def del: o_apply id_apply)
  qed
qed

lemma [code]:
  "add_funs_term_list = Term.fold (const id) (\<lambda>f xs. op # f \<circ> foldr (op \<circ>) xs id)"
    (is "_ = ?fun")
proof (intro ext)
  fix t fs show "add_funs_term_list t fs = ?fun t fs"
  proof (induct t arbitrary: fs)
    case (Var x) show ?case by (simp add: const_def)
  next
    case (Fun f ts fs)
    from foldr_cong[of fs fs ts ts add_funs_term_list, OF _ _ this]
      have IH: "foldr add_funs_term_list ts fs = foldr ?fun ts fs" by blast
    show ?case
      by (simp add: IH const_def foldr_map del: o_apply id_apply)
         simp
  qed
qed

fun prefix_list :: "pos \<Rightarrow> pos list" where
  "prefix_list \<epsilon> = []"
| "prefix_list (PCons i p) = \<epsilon> # map (PCons i) (prefix_list p)"

lemma prefix_list[simp]: "set (prefix_list p) = {q. q < p}"
proof (induction p)
  case Empty
  show ?case by auto
next
  case (PCons i p)
  note IH = this
  show ?case (is "?l = ?r")
  proof (rule set_eqI)
    fix q
    show "q \<in> ?l = (q \<in> ?r)"
    proof (cases q)
      case Empty
      have less: "\<epsilon> < i <# p" unfolding less_pos_def by auto
      show ?thesis unfolding Empty using less by auto
    next
      case (PCons j q')
      show ?thesis unfolding PCons by (auto simp: IH)
    qed
  qed
qed

end
